import pytest

from qcodes.utils.metadata import Metadatable, diff_param_values


class HasSnapshotBase(Metadatable):
    def snapshot_base(self, update=False,
                      params_to_skip_update=None):
        return {'cheese': 'gruyere'}


class HasSnapshot(Metadatable):
    # Users shouldn't do this... but we'll test its behavior
    # for completeness
    def snapshot(self, update=False):
        return {'fruit': 'kiwi'}


DATASETLEFT = {
    "station": {
        "parameters": {
            "apple": {
                "value": "orange"
            }
        },
        "instruments": {
            "correct": {
                "parameters": {
                    "horse": {
                        "value": "battery"
                    },
                    "left": {
                        "value": "only"
                    }
                }
            },
            "another": {
                "parameters": {}
            }
        }
    }
}
DATASETRIGHT = {
    "station": {
        "parameters": {
            "apple": {
                "value": "grape"
            }
        },
        "instruments": {
            "correct": {
                "parameters": {
                    "horse": {
                        "value": "staple"
                    },
                    "right": {
                        "value": "only"
                    }
                }
            },
            "another": {
                "parameters": {
                    "pi": {
                        "value": 3.1
                    }
                }
            }
        }
    }
}


def test_load():
    m = Metadatable()
    assert m.metadata == {}
    m.load_metadata({1: 2, 3: 4})
    assert m.metadata == {1: 2, 3: 4}
    m.load_metadata({1: 5})
    assert m.metadata == {1: 5, 3: 4}


def test_init():
    with pytest.raises(TypeError):
        Metadatable(metadata={2: 3}, not_metadata={4: 5})

    m = Metadatable(metadata={2: 3})
    assert m.metadata == {2: 3}


def test_snapshot():
    m = Metadatable(metadata={6: 7})
    assert m.snapshot_base() == {}
    assert m.snapshot() == {'metadata': {6: 7}}
    del m.metadata[6]
    assert m.snapshot() == {}

    sb = HasSnapshotBase(metadata={7: 8})
    assert sb.snapshot_base() == {'cheese': 'gruyere'}
    assert sb.snapshot() == \
           {'cheese': 'gruyere', 'metadata': {7: 8}}
    del sb.metadata[7]
    assert sb.snapshot() == sb.snapshot_base()

    s = HasSnapshot(metadata={8: 9})
    assert s.snapshot() == {'fruit': 'kiwi'}
    assert s.snapshot_base() == {}
    assert s.metadata == {8: 9}


def test_dataset_diff():
    diff = diff_param_values(DATASETLEFT, DATASETRIGHT)
    assert diff.changed == {
            "apple": ("orange", "grape"),
            ("correct", "horse"): ("battery", "staple"),
        }
    assert diff.left_only == {
            ("correct", "left"): "only"
        }
    assert diff.right_only == {
            ("correct", "right"): "only",
            ("another", "pi"): 3.1,
        }


def test_station_diff():
    left = DATASETLEFT["station"]
    right = DATASETRIGHT["station"]

    diff = diff_param_values(left, right)
    assert diff.changed == {
            "apple": ("orange", "grape"),
            ("correct", "horse"): ("battery", "staple")
        }
    assert diff.left_only == {
            ("correct", "left"): "only"
        }
    assert diff.right_only == {
            ("correct", "right"): "only",
            ("another", "pi"): 3.1,
        }


def test_instrument_diff():
    left = DATASETLEFT["station"]["instruments"]["another"]
    right = DATASETRIGHT["station"]["instruments"]["another"]

    diff = diff_param_values(left, right)

    assert diff.changed == {}
    assert diff.left_only == {}
    assert diff.right_only == {
             "pi": 3.1
        }



from typing import List, Tuple, Dict, Union

import pytest
import numpy as np
from numpy.testing import assert_allclose
import hypothesis.strategies as hst
from hypothesis import given, settings

from qcodes import new_data_set
from qcodes.dataset.measurements import Measurement
from qcodes.dataset.sqlite.connection import atomic_transaction
from qcodes.dataset.descriptions.param_spec import ParamSpecBase
from qcodes.dataset.descriptions.dependencies import InterDependencies_
from qcodes.tests.common import retry_until_does_not_throw

VALUE = Union[str, float, List, np.ndarray, bool]


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
def test_nested_measurement_basic(DAC, DMM, bg_writing):
    meas1 = Measurement()
    meas1.register_parameter(DAC.ch1)
    meas1.register_parameter(DMM.v1, setpoints=(DAC.ch1,))

    meas2 = Measurement()
    meas2.register_parameter(DAC.ch2)
    meas2.register_parameter(DMM.v2, setpoints=(DAC.ch2,))

    with meas1.run(write_in_background=bg_writing) as ds1, \
            meas2.run(write_in_background=bg_writing) as ds2:
        for i in range(10):
            DAC.ch1.set(i)
            DAC.ch2.set(i)
            ds1.add_result((DAC.ch1, i),
                           (DMM.v1, DMM.v1()))
            ds2.add_result((DAC.ch2, i),
                           (DMM.v2, DMM.v2()))

    data1 = ds1.dataset.get_parameter_data()["dummy_dmm_v1"]
    assert len(data1.keys()) == 2
    assert "dummy_dmm_v1" in data1.keys()
    assert "dummy_dac_ch1" in data1.keys()
    assert_allclose(data1["dummy_dmm_v1"], np.zeros(10))
    assert_allclose(data1["dummy_dac_ch1"], np.arange(10))

    data2 = ds2.dataset.get_parameter_data()["dummy_dmm_v2"]
    assert len(data2.keys()) == 2
    assert "dummy_dmm_v2" in data2.keys()
    assert "dummy_dac_ch2" in data2.keys()
    assert_allclose(data2["dummy_dmm_v2"], np.zeros(10))
    assert_allclose(data2["dummy_dac_ch2"], np.arange(10))


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
def test_nested_measurement(bg_writing):
    meas1 = Measurement()
    meas1.register_custom_parameter('foo1')
    meas1.register_custom_parameter('bar1', setpoints=('foo1',))

    meas2 = Measurement()
    meas2.register_custom_parameter('foo2')
    meas2.register_custom_parameter('bar2', setpoints=('foo2',))


    with meas1.run(write_in_background=bg_writing) as ds1, \
            meas2.run(write_in_background=bg_writing) as ds2:
        for i in range(10):
            ds1.add_result(("foo1", i),
                           ("bar1", i**2))
            ds2.add_result(("foo2", 2*i),
                           ("bar2", (2*i)**2))

    data1 = ds1.dataset.get_parameter_data()["bar1"]
    assert len(data1.keys()) == 2
    assert "foo1" in data1.keys()
    assert "bar1" in data1.keys()

    assert_allclose(data1["foo1"], np.arange(10))
    assert_allclose(data1["bar1"], np.arange(10)**2)

    data2 = ds2.dataset.get_parameter_data()["bar2"]
    assert len(data2.keys()) == 2
    assert "foo2" in data2.keys()
    assert "bar2" in data2.keys()
    assert_allclose(data2["foo2"], np.arange(0, 20, 2))
    assert_allclose(data2["bar2"], np.arange(0, 20, 2)**2)


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
@settings(deadline=None, max_examples=25)
@given(outer_len=hst.integers(min_value=1, max_value=100),
       inner_len1=hst.integers(min_value=1, max_value=1000),
       inner_len2=hst.integers(min_value=1, max_value=1000))
def test_nested_measurement_array(bg_writing, outer_len, inner_len1,
                                  inner_len2):
    meas1 = Measurement()
    meas1.register_custom_parameter('foo1', paramtype='numeric')
    meas1.register_custom_parameter('bar1spt', paramtype='array')
    meas1.register_custom_parameter(
        'bar1', setpoints=('foo1', "bar1spt"), paramtype='array'
    )

    meas2 = Measurement()
    meas2.register_custom_parameter('foo2', paramtype='numeric')
    meas2.register_custom_parameter('bar2spt', paramtype='array')
    meas2.register_custom_parameter(
        'bar2', setpoints=('foo2', 'bar2spt',), paramtype='array'
    )

    with meas1.run(write_in_background=bg_writing) as ds1, \
            meas2.run(write_in_background=bg_writing) as ds2:
        for i in range(outer_len):
            bar1sptdata = np.arange(inner_len1)
            bar2sptdata = np.arange(inner_len2)
            ds1.add_result(("foo1", i),
                           ("bar1spt", bar1sptdata),
                           ("bar1", np.ones(inner_len1)*i*bar1sptdata))
            ds2.add_result(("foo2", i),
                           ("bar2spt", bar2sptdata),
                           ("bar2", np.ones(inner_len2)*i*bar2sptdata))

    data1 = ds1.dataset.get_parameter_data()["bar1"]
    assert len(data1.keys()) == 3
    assert "foo1" in data1.keys()
    assert "bar1spt" in data1.keys()
    assert "bar1" in data1.keys()

    expected_foo1_data = np.repeat(np.arange(outer_len),
                                   inner_len1).reshape(outer_len,
                                                       inner_len1)
    expected_bar1spt_data = np.tile(np.arange(inner_len1),
                                    (outer_len, 1))

    assert_allclose(data1["foo1"], expected_foo1_data)
    assert_allclose(data1["bar1spt"], expected_bar1spt_data)
    assert_allclose(data1["bar1"], expected_foo1_data*expected_bar1spt_data)

    data2 = ds2.dataset.get_parameter_data()["bar2"]
    assert len(data2.keys()) == 3
    assert "foo2" in data2.keys()
    assert "bar2spt" in data2.keys()
    assert "bar2" in data2.keys()

    expected_foo2_data = np.repeat(np.arange(outer_len),
                                   inner_len2).reshape(outer_len, inner_len2)
    expected_bar2spt_data = np.tile(np.arange(inner_len2), (outer_len, 1))

    assert_allclose(data2["foo2"], expected_foo2_data)
    assert_allclose(data2["bar2spt"], expected_bar2spt_data)
    assert_allclose(data2["bar2"], expected_foo2_data*expected_bar2spt_data)


@pytest.fixture(scope='function')
def basic_subscriber():
    """
    A basic subscriber that just puts results and length into
    state
    """

    def subscriber(results: List[Tuple[VALUE]], length: int,
                   state: Dict) -> None:
        state[length] = results

    return subscriber


@pytest.mark.flaky(reruns=5)
@pytest.mark.serial
def test_subscription_on_dual_datasets(experiment, basic_subscriber):
    xparam = ParamSpecBase(name='x',
                           paramtype='numeric',
                           label='x parameter',
                           unit='V')
    yparam1 = ParamSpecBase(name='y1',
                            paramtype='numeric',
                            label='y parameter',
                            unit='Hz')
    yparam2 = ParamSpecBase(name='y2',
                            paramtype='numeric',
                            label='y parameter',
                            unit='Hz')

    dataset1 = new_data_set("test-dataset-1")
    idps_1 = InterDependencies_(dependencies={yparam1: (xparam,)})
    dataset1.set_interdependencies(idps_1)
    dataset1.mark_started()

    sub_id_1 = dataset1.subscribe(basic_subscriber, min_wait=0, min_count=1,
                                  state={})

    assert len(dataset1.subscribers) == 1
    assert list(dataset1.subscribers.keys()) == [sub_id_1]

    dataset2 = new_data_set("test-dataset-2")
    idps_2 = InterDependencies_(dependencies={yparam2: (xparam,)})
    dataset2.set_interdependencies(idps_2)
    dataset2.mark_started()

    sub_id_2 = dataset2.subscribe(basic_subscriber, min_wait=0, min_count=1,
                                  state={})

    assert len(dataset2.subscribers) == 1
    assert list(dataset2.subscribers.keys()) == [sub_id_2]

    assert sub_id_1 != sub_id_2


    expected_state_1 = {}
    expected_state_2 = {}

    for x in range(10):
        y1 = -x**2
        y2 = x ** 2
        dataset1.add_results([{'x': x, 'y1': y1}])
        dataset2.add_results([{'x': x, 'y2': y2}])
        expected_state_1[x+1] = [(x, y1)]
        expected_state_2[x + 1] = [(x, y2)]

        @retry_until_does_not_throw(
            exception_class_to_expect=AssertionError, delay=0.5, tries=10)
        def assert_expected_state():
            assert dataset1.subscribers[sub_id_1].state == expected_state_1
            assert dataset2.subscribers[sub_id_2].state == expected_state_2

        assert_expected_state()

    dataset1.unsubscribe(sub_id_1)
    dataset2.unsubscribe(sub_id_2)

    assert len(dataset1.subscribers) == 0
    assert list(dataset1.subscribers.keys()) == []

    assert len(dataset2.subscribers) == 0
    assert list(dataset2.subscribers.keys()) == []

    # Ensure the trigger for the subscriber has been removed from the database
    get_triggers_sql = "SELECT * FROM sqlite_master WHERE TYPE = 'trigger';"
    triggers1 = atomic_transaction(
        dataset1.conn, get_triggers_sql).fetchall()
    assert len(triggers1) == 0

    triggers2 = atomic_transaction(
        dataset2.conn, get_triggers_sql).fetchall()
    assert len(triggers2) == 0



from copy import deepcopy
from typing import Any, TYPE_CHECKING, List, Optional, Union, Iterator, Dict, \
    Sequence

from qcodes.utils.helpers import (is_sequence, permissive_range, make_sweep,
                                  named_repr)
from qcodes.utils.metadata import Metadatable
if TYPE_CHECKING:
    from qcodes.instrument.parameter import _BaseParameter


class SweepValues(Metadatable):
    """
    Base class for sweeping a parameter.

    Must be subclassed to provide the sweep values
    Intended use is to iterate over in a sweep, so it must support:

    >>> .__iter__ # (and .__next__ if necessary).
    >>> .set # is provided by the base class

    Optionally, it can have a feedback method that allows the sweep to pass
    measurements back to this object for adaptive sampling:

    >>> .feedback(set_values, measured_values)

    Todo:
        - Link to adawptive sweep

    Args:
        parameter: the target of the sweep, an object with
         set, and optionally validate methods

        **kwargs: Passed on to Metadatable parent

    Raises:
        TypeError: when parameter is not settable

    See AdaptiveSweep for an example

    example usage:

    >>> for i, value in eumerate(sv):
            sv.set(value)
            sleep(delay)
            vals = measure()
            sv.feedback((i, ), vals) # optional - sweep should not assume
                                     # .feedback exists

    note though that sweeps should only require set and __iter__ - ie
    "for val in sv", so any class that implements these may be used in sweeps.

    That allows things like adaptive sampling, where you don't know ahead of
    time what the values will be or even how many there are.
    """
    def __init__(self, parameter: '_BaseParameter', **kwargs: Any):
        super().__init__(**kwargs)
        self.parameter = parameter
        self.name = parameter.name
        self._values: List[Any] = []

        # allow has_set=False to override the existence of a set method,
        # but don't require it to be present (and truthy) otherwise
        if not (getattr(parameter, 'set', None) and
                getattr(parameter, 'has_set', True)):
            raise TypeError(f'parameter {parameter} is not settable')

        self.set = parameter.set

    def validate(self, values: Sequence[Any]) -> None:
        """
        Check that all values are allowed for this Parameter.

        Args:
            values: values to be validated.
        """
        if hasattr(self.parameter, 'validate'):
            for value in values:
                self.parameter.validate(value)

    def __iter__(self) -> Iterator[Any]:
        """
        must be overridden (along with __next__ if this returns self)
        by a subclass to tell how to iterate over these values
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return named_repr(self)


class SweepFixedValues(SweepValues):
    """
    A fixed collection of parameter values to be iterated over during a sweep.

    Args:
        parameter: the target of the sweep, an object with set and
            optionally validate methods

        keys: one or a sequence of items, each of which can be:
            - a single parameter value
            - a sequence of parameter values
            - a slice object, which MUST include all three args

        start: The starting value of the sequence.
        stop: The end value of the sequence.
        step:  Spacing between values.
        num: Number of values to generate.


    A SweepFixedValues object is normally created by slicing a Parameter p:

    >>>  sv = p[1.2:2:0.01]  # slice notation
    sv = p[1, 1.1, 1.3, 1.6]  # explicit individual values
    sv = p[1.2:2:0.01, 2:3:0.02]  # sequence of slices
    sv = p[logrange(1,10,.01)]  # some function that returns a sequence

    You can also use list operations to modify these:

    >>> sv += p[2:3:.01] # (another SweepFixedValues of the same parameter)
    sv += [4, 5, 6] # (a bare sequence)
    sv.extend(p[2:3:.01])
    sv.append(3.2)
    sv.reverse()
    sv2 = reversed(sv)
    sv3 = sv + sv2
    sv4 = sv.copy()

    note though that sweeps should only require set and __iter__ - ie
    "for val in sv", so any class that implements these may be used in sweeps.
    That allows things like adaptive sampling, where you don't know ahead of
    time what the values will be or even how many there are.
    """
    def __init__(self, parameter: '_BaseParameter',
                 keys: Optional[Any] = None,
                 start: Optional[float] = None,
                 stop: Optional[float] = None,
                 step: Optional[float] = None,
                 num: Optional[int] = None):
        super().__init__(parameter)
        self._snapshot: Dict[str, Any] = {}
        self._value_snapshot: List[Dict[str, Any]] = []

        if keys is None:
            if start is None:
                raise ValueError('If keys is None, start needs to be not None.')
            if stop is None:
                raise ValueError('If keys is None, stop needs to be not None.')
            keys = make_sweep(start=start, stop=stop,
                              step=step, num=num)
            self._values = keys
            self._add_linear_snapshot(self._values)

        elif isinstance(keys, slice):
            self._add_slice(keys)
            self._add_linear_snapshot(self._values)

        elif is_sequence(keys):
            for key in keys:
                if isinstance(key, slice):
                    self._add_slice(key)
                elif is_sequence(key):
                    # not sure if we really need to support this (and I'm not
                    # going to recurse any more!) but we will get nested lists
                    # if for example someone does `p[list1, list2]`
                    self._values.extend(key)
                else:
                    # assume a single value
                    self._values.append(key)
            # we dont want the snapshot to go crazy on big data
            if self._values:
                self._add_sequence_snapshot(self._values)

        else:
            # assume a single value
            self._values.append(keys)
            self._value_snapshot.append({'item': keys})

        self.validate(self._values)

    def _add_linear_snapshot(self, vals: List[Any]) -> None:
        self._value_snapshot.append({'first': vals[0],
                                     'last': vals[-1],
                                     'num': len(vals),
                                     'type': 'linear'})

    def _add_sequence_snapshot(self, vals: Sequence[Any]) -> None:
        self._value_snapshot.append({'min': min(vals),
                                     'max': max(vals),
                                     'first': vals[0],
                                     'last': vals[-1],
                                     'num': len(vals),
                                     'type': 'sequence'})

    def _add_slice(self, slice_: slice) -> None:
        if slice_.start is None or slice_.stop is None or slice_.step is None:
            raise TypeError('all 3 slice parameters are required, ' +
                            f'{slice_} is missing some')
        p_range = permissive_range(slice_.start, slice_.stop, slice_.step)
        self._values.extend(p_range)

    def append(self, value: Any) -> None:
        """
        Append a value.

        Args:
            value: new value to append
        """
        self.validate((value,))
        self._values.append(value)
        self._value_snapshot.append({'item': value})

    def extend(
            self,
            new_values: Union[Sequence[Any], 'SweepFixedValues']
    ) -> None:
        """
        Extend sweep with new_values

        Args:
            new_values: new values to append

        Raises:
            TypeError: if new_values is not Sequence, nor SweepFixedValues
        """
        if isinstance(new_values, SweepFixedValues):
            if new_values.parameter is not self.parameter:
                raise TypeError(
                    'can only extend SweepFixedValues of the same parameters')
            # these values are already validated
            self._values.extend(new_values._values)
            self._value_snapshot.extend(new_values._value_snapshot)
        elif is_sequence(new_values):
            self.validate(new_values)
            self._values.extend(new_values)
            self._add_sequence_snapshot(new_values)
        else:
            raise TypeError(
                f'cannot extend SweepFixedValues with {new_values}')

    def copy(self) -> 'SweepFixedValues':
        """
        Copy this SweepFixedValues.

        Returns:
            SweepFixedValues of copied values
        """
        new_sv = SweepFixedValues(self.parameter, [])
        # skip validation by adding values and snapshot separately
        # instead of on init
        new_sv._values = self._values[:]
        new_sv._value_snapshot = deepcopy(self._value_snapshot)
        return new_sv

    def reverse(self) -> None:
        """ Reverse SweepFixedValues in place. """
        self._values.reverse()
        self._value_snapshot.reverse()
        for snap in self._value_snapshot:
            if 'first' in snap and 'last' in snap:
                snap['last'], snap['first'] = snap['first'], snap['last']

    def snapshot_base(self, update: Optional[bool] = False,
                      params_to_skip_update: Optional[Sequence[str]] = None
                      ) -> Dict[Any, Any]:
        """
        Snapshot state of SweepValues.

        Args:
            update: Place holder for API compatibility.
            params_to_skip_update: Place holder for API compatibility.

        Returns:
            dict: base snapshot
        """
        self._snapshot['parameter'] = self.parameter.snapshot(update=update)
        self._snapshot['values'] = self._value_snapshot
        return self._snapshot

    def __iter__(self) -> Iterator[Any]:
        return iter(self._values)

    def __getitem__(self, key: slice) -> Any:
        return self._values[key]

    def __len__(self) -> int:
        return len(self._values)

    def __add__(self, other: Union[Sequence[Any], 'SweepFixedValues']
                ) -> 'SweepFixedValues':
        new_sv = self.copy()
        new_sv.extend(other)
        return new_sv

    def __iadd__(self, values: Union[Sequence[Any], 'SweepFixedValues']
                 ) -> 'SweepFixedValues':
        self.extend(values)
        return self

    def __contains__(self, value: float) -> bool:
        return value in self._values

    def __reversed__(self) -> 'SweepFixedValues':
        new_sv = self.copy()
        new_sv.reverse()
        return new_sv



"""This file contains functions to displays an interactive widget
with information about `qcodes.experiments()`."""

from __future__ import annotations

import io
import operator
import traceback
from datetime import datetime
from functools import partial, reduce
from typing import (Any, Callable, Dict, Optional, Sequence, Iterable,
    TYPE_CHECKING, Union, List)

import matplotlib.pyplot as plt
from IPython.core.display import display
from IPython.display import clear_output
from ipywidgets import (
    HTML,
    Box,
    Button,
    GridspecLayout,
    HBox,
    Label,
    Layout,
    Output,
    Tab,
    Textarea,
    VBox,
)
from ruamel.yaml import YAML
from typing_extensions import Literal

import qcodes
from qcodes.dataset import initialise_or_create_database_at
from qcodes.dataset.data_set import DataSet
from qcodes.dataset.plotting import plot_dataset

if TYPE_CHECKING:
    from qcodes.dataset.descriptions.param_spec import ParamSpecBase

_META_DATA_KEY = "widget_notes"


def _get_in(nested_keys: Sequence[str], dct: Dict[str, Any]) -> Dict[str, Any]:
    """ Returns dct[i0][i1]...[iX] where [i0, i1, ..., iX]==nested_keys."""
    return reduce(operator.getitem, nested_keys, dct)


def button(
    description: str,
    button_style: Optional[str] = None,
    on_click: Optional[Callable[[Any], None]] = None,
    tooltip: Optional[str] = None,
    layout_kwargs: Optional[Dict[str, Any]] = None,
    button_kwargs: Optional[Dict[str, Any]] = None,
) -> Button:
    """Returns a `ipywidgets.Button`."""
    layout_kwargs = layout_kwargs or {}
    but = Button(
        description=description,
        button_style=button_style,
        layout=Layout(
            height=layout_kwargs.pop("height", "auto"),
            width=layout_kwargs.pop("width", "auto"),
            **layout_kwargs,
        ),
        tooltip=tooltip or description,
        **(button_kwargs or {}),
    )
    if on_click is not None:
        but.on_click(on_click)
    return but


def button_to_text(title: str, body: str) -> Box:
    def _button_to_input(
        title: str, body: str, box: Box
    ) -> Callable[[Button], None]:
        def on_click(_: Button) -> None:
            text_input = Textarea(
                value=body,
                placeholder="Enter text",
                disabled=True,
                layout=Layout(height="300px", width="auto"),
            )
            back_button = button(
                "Back",
                "warning",
                on_click=_back_button(title, body, box),
                button_kwargs=dict(icon="undo"),
            )
            box.children = (text_input, back_button)

        return on_click

    def _back_button(
        title: str, body: str, box: Box
    ) -> Callable[[Button], None]:
        def on_click(_: Button) -> None:
            box.children = (_changeable_button(title, body, box),)

        return on_click

    def _changeable_button(title: str, body: str, box: Box) -> Button:
        return button(
            title, "success", on_click=_button_to_input(title, body, box),
        )

    box = VBox([], layout=Layout(height="auto", width="auto"))
    box.children = (_changeable_button(title, body, box),)
    return box


def label(description: str) -> Label:
    """Returns a `ipywidgets.Label` with text."""
    return Label(
        value=description, layout=Layout(height="max-content", width="auto")
    )


def _update_nested_dict_browser(
    nested_keys: Sequence[str], nested_dict: Dict[Any, Any], box: Box
) -> Callable[[Button], None]:
    def update_box(_: Button) -> None:
        box.children = (_nested_dict_browser(nested_keys, nested_dict, box),)

    return update_box


def _nested_dict_browser(
    nested_keys: Sequence[str],
    nested_dict: Dict[Any, Any],
    box: Box,
    max_nrows: int = 30,
) -> GridspecLayout:
    """Generates a `GridspecLayout` of the ``nested_keys`` in ``nested_dict``
    which is put inside of ``box``.

    Args:
        nested_keys: A sequence of keys of the nested dict. e.g., if
            ``nested_keys=['a', 'b']`` then ``nested_dict['a']['b']``.
        nested_dict: A dictionary that can contain more dictionaries as keys.
        box: An `ipywidgets.Box` instance.
        max_nrows: The maximum number of rows that can be displayed at once.
            Whenever the table has less than ``max_nrows`` rows, the table is
            displayed in 3 columns, otherwise it's 2 columns.
    """

    def _should_expand(x: Any) -> bool:
        return isinstance(x, dict) and x != {}

    col_widths = [8, 16, 30]
    selected_table = _get_in(nested_keys, nested_dict)
    nrows = (
        sum(
            len(v) if _should_expand(v) else 1 for v in selected_table.values()
        )
        + 1
    )
    ncols = 3

    if nrows > max_nrows:
        nrows = len(selected_table) + 1
        col_widths.pop(1)
        ncols = 2

    grid = GridspecLayout(nrows, col_widths[-1])
    update = partial(
        _update_nested_dict_browser, nested_dict=nested_dict, box=box
    )

    # Header
    title = " ► ".join(nested_keys)
    grid[0, :-1] = button(title, "success")
    up_click = update(nested_keys[:-1])
    grid[0, -1] = button("↰", "info", up_click)

    # Body

    row_index = 1
    for k, v in selected_table.items():
        row_length = len(v) if _should_expand(v) and ncols == 3 else 1
        but = button(k, "info", up_click)
        grid[row_index : row_index + row_length, : col_widths[0]] = but
        if _should_expand(v):
            if ncols == 3:
                for k_, v_ in v.items():
                    but = button(k_, "danger", update([*nested_keys, k]))
                    grid[row_index, col_widths[0] : col_widths[1]] = but
                    if _should_expand(v_):
                        sub_keys = ", ".join(v_.keys())
                        but = button(
                            sub_keys, "warning", update([*nested_keys, k, k_])
                        )
                    else:
                        but = label(str(v_))
                    grid[row_index, col_widths[1] :] = but
                    row_index += 1
            else:
                sub_keys = ", ".join(v.keys())
                grid[row_index, col_widths[0] :] = button(
                    sub_keys, "danger", update([*nested_keys, k])
                )
                row_index += 1
        else:
            grid[row_index, col_widths[0] :] = label(str(v))
            row_index += 1
    return grid


def nested_dict_browser(
    nested_dict: Dict[Any, Any], nested_keys: Sequence[str] = ()
) -> Box:
    """Returns a widget to interactive browse a nested dictionary."""
    box = Box([])
    _update_nested_dict_browser(nested_keys, nested_dict, box)(None)
    return box


def _plot_ds(ds: DataSet) -> None:
    plot_dataset(ds)  # might fail
    plt.show()


def _do_in_tab(tab: Tab, ds: DataSet, which: Literal["plot", "snapshot"]
               ) -> Callable[[Button], None]:
    """Performs an operation inside of a subtab of a `ipywidgets.Tab`.

    Args
        tab: Instance of `ipywidgets.Tab`.
        ds: A qcodes.DataSet instance.
        which: Either "plot" or "snapshot".
    """
    assert which in ("plot", "snapshot")

    def delete_tab(output: Output, tab: Tab) -> Callable[[Button], None]:
        def on_click(_: Button) -> None:
            tab.children = tuple(c for c in tab.children if c != output)

        return on_click

    def _on_click(_: Button) -> None:
        assert which in ("plot", "snapshot")
        title = f"RID #{ds.captured_run_id} {which}"
        i = next(
            (i for i in range(len(tab.children)) if tab.get_title(i) == title),
            None,
        )
        if i is not None:
            # Plot/snapshot is already in the tab
            tab.selected_index = i
            return
        out = Output()
        tab.children += (out,)
        i = len(tab.children) - 1
        tab.set_title(i, title)
        with out:
            clear_output(wait=True)

            close_button = button(
                f"Close {which}",
                "danger",
                on_click=delete_tab(out, tab),
                button_kwargs=dict(icon="eraser"),
            )
            display(close_button)

            try:
                if which == "plot":
                    _plot_ds(ds)
                elif which == "snapshot":
                    snapshot = ds.snapshot
                    if snapshot is not None:
                        display(nested_dict_browser(snapshot))
                    else:
                        print("This dataset has no snapshot")
            except Exception:
                traceback.print_exc()
        tab.selected_index = i

    return _on_click


def create_tab(do_display: bool = True) -> Tab:
    """Creates a `ipywidgets.Tab` which can display outputs in its tabs."""
    output = Output()
    tab = Tab(children=(output,))

    tab.set_title(0, "Info")
    if do_display:
        display(tab)

    with output:
        # Prints it in the Output inside the tab.
        print("Plots and snapshots will show up here!")
    return tab


def editable_metadata(ds: DataSet) -> Box:
    def _button_to_input(text: str, box: Box) -> Callable[[Button], None]:
        def on_click(_: Button) -> None:
            text_input = Textarea(
                value=text,
                placeholder="Enter text",
                disabled=False,
                layout=Layout(height="auto", width="auto"),
            )
            save_button = button(
                "",
                "success",
                on_click=_save_button(box, ds),
                button_kwargs=dict(icon="save"),
                layout_kwargs=dict(width="50%"),
            )
            cancel_button = button(
                "",
                "danger",
                on_click=_save_button(box, ds, do_save=False),
                button_kwargs=dict(icon="close"),
                layout_kwargs=dict(width="50%"),
            )
            subbox = HBox([save_button, cancel_button],)
            box.children = (text_input, subbox)

        return on_click

    def _save_button(
        box: Box, ds: DataSet, do_save: bool = True
    ) -> Callable[[Button], None]:
        def on_click(_: Button) -> None:
            text = box.children[0].value
            if do_save:
                ds.add_metadata(tag=_META_DATA_KEY, metadata=text)
            box.children = (_changeable_button(text, box),)

        return on_click

    def _changeable_button(text: str, box: Box) -> Button:
        return button(
            text,
            "success",
            on_click=_button_to_input(text, box),
            button_kwargs=dict(icon="edit") if text == "" else {},
        )

    text = ds.metadata.get(_META_DATA_KEY, "")
    box = VBox([], layout=Layout(height="auto", width="auto"))
    box.children = (_changeable_button(text, box),)
    return box


def _yaml_dump(dct: Dict[str, Any]) -> str:
    with io.StringIO() as f:
        YAML().dump(dct, f)
        return f.getvalue()


def _get_parameters(ds: DataSet) -> Dict[str, Dict[str, Any]]:
    independent = {}
    dependent = {}

    def _get_attr(p: ParamSpecBase) -> Dict[str, Any]:
        return {
            "unit": p.unit,
            "label": p.label,
            "type": p.type,
        }

    for dep, indeps in ds.description.interdeps.dependencies.items():
        attrs = _get_attr(dep)
        attrs["depends_on"] = [p.name for p in indeps]
        dependent[dep.name] = attrs
        for p in indeps:
            independent[p.name] = _get_attr(p)
    return {"independent": independent, "dependent": dependent}


def _get_experiment_button(ds: DataSet) -> Box:
    title = f"{ds.exp_name}, {ds.sample_name}"
    body = _yaml_dump(
        {
            ".exp_name": ds.exp_name,
            ".sample_name": ds.sample_name,
            ".name": ds.name,
            ".path_to_db": ds.path_to_db,
            ".exp_id": ds.exp_id,
        }
    )
    return button_to_text(title, body)


def _get_timestamp_button(ds: DataSet) -> Box:
    try:
        total_time = str(
            datetime.fromtimestamp(ds.run_timestamp_raw)  # type: ignore
            - datetime.fromtimestamp(ds.completed_timestamp_raw)  # type: ignore
        )
    except TypeError:
        total_time = "?"
    start = ds.run_timestamp()
    body = _yaml_dump(
        {
            ".run_timestamp": start,
            ".completed_timestamp": ds.completed_timestamp(),
            "total_time": total_time,
        }
    )
    return button_to_text(start or "", body)


def _get_run_id_button(ds: DataSet) -> Box:
    title = str(ds.run_id)
    body = _yaml_dump(
        {
            ".guid": ds.guid,
            ".captured_run_id": ds.captured_run_id,
            ".run_id": ds.run_id,
        }
    )
    return button_to_text(title, body)


def _get_parameters_button(ds: DataSet) -> VBox:
    parameters = _get_parameters(ds)
    title = ds.parameters or ""
    return button_to_text(title, _yaml_dump(parameters))


def _get_snapshot_button(ds: DataSet, tab: Tab) -> Button:
    return button(
        "",
        "warning",
        tooltip="Click to open this DataSet's snapshot in a tab above.",
        on_click=_do_in_tab(tab, ds, "snapshot"),
        button_kwargs=dict(icon="camera"),
    )


def _get_plot_button(ds: DataSet, tab: Tab) -> Button:
    return button(
        "",
        "warning",
        tooltip="Click to open this DataSet's plot in a tab above.",
        on_click=_do_in_tab(tab, ds, "plot"),
        button_kwargs=dict(icon="line-chart"),
    )


def _experiment_widget(data_sets: Iterable[DataSet], tab: Tab
                       ) -> GridspecLayout:
    """Show a `ipywidgets.GridspecLayout` with information about the
    loaded experiment. The clickable buttons can perform an action in ``tab``.
    """
    header_names = [
        "Run ID",
        "Experiment",
        "Name",
        "Parameters",
        "MSMT Time",
        "Notes",
        "Snapshot",
        "Plot",
    ]

    header = {n: button(n, "info") for n in header_names}
    rows = [header]
    for ds in data_sets:
        row = {}
        row["Run ID"] = _get_run_id_button(ds)
        row["Experiment"] = _get_experiment_button(ds)
        row["Name"] = label(ds.name)
        row["Notes"] = editable_metadata(ds)
        row["Parameters"] = _get_parameters_button(ds)
        row["MSMT Time"] = _get_timestamp_button(ds)
        row["Snapshot"] = _get_snapshot_button(ds, tab)
        row["Plot"] = _get_plot_button(ds, tab)
        rows.append(row)

    grid = GridspecLayout(n_rows=len(rows), n_columns=len(header_names))

    empty_label = label("")
    for i, row in enumerate(rows):
        for j, name in enumerate(header_names):
            grid[i, j] = row.get(name, empty_label)

    grid.layout.grid_template_rows = "auto " * len(rows)
    grid.layout.grid_template_columns = "auto " * len(header_names)
    return grid


def experiments_widget(
    db: Optional[str] = None,
    data_sets: Optional[Sequence[DataSet]] = None,
    *,
    sort_by: Optional[Literal["timestamp", "run_id"]] = "run_id",
) -> VBox:
    r"""Displays an interactive widget that shows the ``qcodes.experiments()``.

    With the edit button in the column "Notes" one can make persistent
    changes to the `~qcodes.dataset.data_set.DataSet`\s attribute
    ``metadata`` in the key "widget_notes".
    Expanding the coordinates or variables buttons, reveals more options, such
    as plotting or the ability to easily browse
    the `~qcodes.dataset.data_set.DataSet`\s snapshot.

    Args
        db: Optionally pass a database file, if no database has been loaded.
        data_sets: Sequence of `~qcodes.dataset.data_set.DataSet`s.
            If datasets are explicitly provided via this argument, the ``db``
            argument has no effect.
        sort_by: Sort datasets in widget by either "timestamp" (newest first),
            "run_id" or None (no predefined sorting).
    """
    if data_sets is None:
        if db is not None:
            initialise_or_create_database_at(db)
        data_sets = [
            ds for exp in qcodes.experiments() for ds in exp.data_sets()
        ]
    if sort_by == "run_id":
        data_sets = sorted(data_sets, key=lambda ds: ds.run_id)
    elif sort_by == "timestamp":
        data_sets = sorted(
            data_sets,
            key=lambda ds: ds.run_timestamp_raw if ds.run_timestamp_raw is not None else 0,
            reverse=True
        )

    title = HTML("<h1>QCoDeS experiments widget</h1>")
    tab = create_tab(do_display=False)
    grid = _experiment_widget(data_sets, tab)
    return VBox([title, tab, grid])



import numpy as np
import pytest
from qcodes.utils.helpers import compare_dictionaries


def test_same():
    # NOTE(alexcjohnson): the numpy array and list compare equal,
    # even though a list and tuple would not. See TODO in
    # compare_dictionaries.
    a = {'a': 1, 2: [3, 4, {5: 6}], 'b': {'c': 'd'}, 'x': np.array([7, 8])}
    b = {'a': 1, 2: [3, 4, {5: 6}], 'b': {'c': 'd'}, 'x': [7, 8]}

    match, err = compare_dictionaries(a, b)
    assert match
    assert err == ''


def test_bad_dict():
    # NOTE(alexcjohnson):
    # this is a valid dict, but not good JSON because the tuple key cannot
    # be converted into a string.
    # It throws an error in compare_dictionaries, which is likely what we
    # want, but we should be aware of it.
    a = {(5, 6): (7, 8)}
    with pytest.raises(TypeError):
        compare_dictionaries(a, a)


def test_key_diff():
    a = {'a': 1, 'c': 4}
    b = {'b': 1, 'c': 4}

    match, err = compare_dictionaries(a, b)

    assert not match
    assert 'Key d1[a] not in d2' in err
    assert 'Key d2[b] not in d1' in err

    # try again with dict names for completeness
    match, err = compare_dictionaries(a, b, 'a', 'b')

    assert not match
    assert 'Key a[a] not in b' in err
    assert 'Key b[b] not in a' in err


def test_val_diff_simple():
    a = {'a': 1}
    b = {'a': 2}

    match, err = compare_dictionaries(a, b)

    assert not match
    assert 'Value of "d1[a]" ("1", type"<class \'int\'>") not same as' in err
    assert '"d2[a]" ("2", type"<class \'int\'>")' in err


def test_val_diff_seq():
    # NOTE(alexcjohnson):
    # we don't dive recursively into lists at the moment.
    # Perhaps we want to? Seems like list equality does a deep comparison,
    # so it's not necessary to get ``match`` right, but the error message
    # could be more helpful if we did.
    a = {'a': [1, {2: 3}, 4]}
    b = {'a': [1, {5: 6}, 4]}

    match, err = compare_dictionaries(a, b)

    assert not match
    assert 'Value of "d1[a]" ("[1, {2: 3}, 4]", ' \
           'type"<class \'list\'>") not same' in err
    assert '"d2[a]" ("[1, {5: 6}, 4]", type"<class \'list\'>")' in \
           err


def test_nested_key_diff():
    a = {'a': {'b': 'c'}}
    b = {'a': {'d': 'c'}}

    match, err = compare_dictionaries(a, b)

    assert not match
    assert 'Key d1[a][b] not in d2' in err
    assert 'Key d2[a][d] not in d1' in err



"""
Test attribute_set_to context manager
"""
from qcodes.utils.helpers import attribute_set_to


class X:
    y = 0


def test_attribute_set_to_value():
    """Test setting attribute to a value"""
    x = X()
    x.y = 1

    assert 1 == x.y

    with attribute_set_to(x, 'y', 2):
        assert 2 == x.y

    assert 1 == x.y


def test_attribute_set_to_object():
    """Test setting attribute to an object"""
    x = X()
    original_object = X()
    x.y = original_object

    assert original_object == x.y
    assert original_object is x.y

    new_object = X()
    with attribute_set_to(x, 'y', new_object):
        assert new_object == x.y
        assert new_object is x.y

    assert original_object == x.y
    assert original_object is x.y



from typing import Generator

import pytest

from qcodes.dataset.export_config import DataExportType
from qcodes.dataset.exporters.export_info import ExportInfo


@pytest.fixture(name="basic_export_info")
def _make_basic_export_info() -> Generator[ExportInfo, None, None]:
    nc_path = "D:\\data\\33.nc"
    csv_path = "D:\\data\\33.csv"

    export_info = ExportInfo(
        {DataExportType.NETCDF.value: nc_path, DataExportType.CSV.value: csv_path}
    )
    yield export_info


def test_export_info_basic() -> None:
    nc_path = "D:\\data\\33.nc"
    csv_path = "D:\\data\\33.csv"

    a = ExportInfo(
        {DataExportType.NETCDF.value: nc_path, DataExportType.CSV.value: csv_path}
    )
    assert a.export_paths[DataExportType.NETCDF.value] == nc_path
    assert a.export_paths[DataExportType.CSV.value] == csv_path


def test_invalid_key_raises() -> None:
    nd_path = "D:\\data\\33.nd"

    with pytest.warns(Warning, match="The supported export types are"):
        _ = ExportInfo({"nd": nd_path})


def test_export_info_json_roundtrip(basic_export_info) -> None:

    exported_str = basic_export_info.to_str()
    loaded_export_info = ExportInfo.from_str(exported_str)
    assert loaded_export_info == basic_export_info


def test_init_from_empty_str() -> None:
    export_info = ExportInfo.from_str("")
    assert export_info.export_paths == {}



import pytest

from qcodes.instrument.parameter import Parameter, _BaseParameter
from .conftest import (OverwriteGetParam, OverwriteSetParam,
                       GetSetRawParameter, ParameterMemory)


def test_parameter_with_overwritten_get_raises():
    """
    Test that creating a parameter that overwrites get and set raises runtime errors
    """

    with pytest.raises(RuntimeError) as record:
        OverwriteGetParam(name='foo')
    assert "Overwriting get in a subclass of _BaseParameter: foo is not allowed." == str(record.value)


def test_parameter_with_overwritten_set_raises():
    """
    Test that creating a parameter that overwrites get and set raises runtime errors
    """
    with pytest.raises(RuntimeError) as record:
        OverwriteSetParam(name='foo')
    assert "Overwriting set in a subclass of _BaseParameter: foo is not allowed." == str(record.value)


@pytest.mark.parametrize("get_cmd, set_cmd", [(False, False), (False, None), (None, None), (None, False),
                                              (lambda: 1, lambda x: x)])
def test_gettable_settable_attributes_with_get_set_cmd(get_cmd, set_cmd):
    a = Parameter(name='foo',
                  get_cmd=get_cmd,
                  set_cmd=set_cmd)
    expected_gettable = get_cmd is not False
    expected_settable = set_cmd is not False

    assert a.gettable is expected_gettable
    assert a.settable is expected_settable


@pytest.mark.parametrize("baseclass", [_BaseParameter, Parameter])
def test_gettable_settable_attributes_with_get_set_raw(baseclass):
    """Test that parameters that have get_raw,set_raw are
    listed as gettable/settable and reverse."""

    class GetSetParam(baseclass):
        def __init__(self, *args, initial_value=None, **kwargs):
            self._value = initial_value
            super().__init__(*args, **kwargs)

        def get_raw(self):
            return self._value

        def set_raw(self, value):
            self._value = value

    a = GetSetParam('foo', instrument=None, initial_value=1)

    assert a.gettable is True
    assert a.settable is True

    b = _BaseParameter('foo', None)

    assert b.gettable is False
    assert b.settable is False


@pytest.mark.parametrize("working_get_cmd", (False, None))
@pytest.mark.parametrize("working_set_cmd", (False, None))
def test_get_raw_and_get_cmd_raises(working_get_cmd, working_set_cmd):
    with pytest.raises(TypeError, match="get_raw"):
        GetSetRawParameter(name="param1", get_cmd="GiveMeTheValue", set_cmd=working_set_cmd)
    with pytest.raises(TypeError, match="set_raw"):
        GetSetRawParameter(name="param2", set_cmd="HereIsTheValue {}", get_cmd=working_get_cmd)
    GetSetRawParameter("param3", get_cmd=working_get_cmd, set_cmd=working_set_cmd)


def test_get_on_parameter_marked_as_non_gettable_raises():
    a = Parameter("param")
    a._gettable = False
    with pytest.raises(TypeError, match="Trying to get a parameter that is not gettable."):
        a.get()


def test_set_on_parameter_marked_as_non_settable_raises():
    a = Parameter("param", set_cmd=None)
    a.set(2)
    assert a.get() == 2
    a._settable = False
    with pytest.raises(TypeError, match="Trying to set a parameter that is not settable."):
        a.set(1)
    assert a.get() == 2


def test_settable():
    mem = ParameterMemory()

    p = Parameter('p', set_cmd=mem.set, get_cmd=False)

    p(10)
    assert mem.get() == 10
    with pytest.raises(NotImplementedError):
        p()

    assert hasattr(p, 'set')
    assert p.settable
    assert not hasattr(p, 'get')
    assert not p.gettable

    # For settable-only parameters, using ``cache.set`` may not make
    # sense, nevertheless, it works
    p.cache.set(7)
    assert p.get_latest() == 7


def test_gettable():
    mem = ParameterMemory()
    p = Parameter('p', get_cmd=mem.get)
    mem.set(21)

    assert p() == 21
    assert p.get() == 21

    with pytest.raises(NotImplementedError):
        p(10)

    assert hasattr(p, 'get')
    assert p.gettable
    assert not hasattr(p, 'set')
    assert not p.settable

    p.cache.set(7)
    assert p.get_latest() == 7
    # Nothing has been passed to the "instrument" at ``cache.set``
    # call, hence the following assertions should hold
    assert mem.get() == 21
    assert p() == 21
    assert p.get_latest() == 21



import ctypes
import logging
import time
import warnings
import sys
from typing import List, Dict, Union, Sequence, Optional, Any, Iterator, cast, TypeVar, Type, Generic
from contextlib import contextmanager

import numpy as np

from qcodes.instrument.base import Instrument
from .ats_api import AlazarATSAPI
from .utils import TraceParameter
from .helpers import CapabilityHelper
from .constants import NUMBER_OF_CHANNELS_FROM_BYTE_REPR, max_buffer_size


logger = logging.getLogger(__name__)

OutputType = TypeVar('OutputType')

CtypesTypes = Union[Type[ctypes.c_uint8],
                    Type[ctypes.c_uint16],
                    Type[ctypes.c_uint32],
                    Type[ctypes.c_int32],
                    Type[ctypes.c_float]]


class AlazarTech_ATS(Instrument):
    """
    This is the qcodes driver for Alazar data acquisition cards

    status: beta-version

    Args:
        name: name for this instrument
        system_id: target system id for this board
        board_id: target board id within the system for this board
        dll_path: path to the ATS driver dll library file
        api: AlazarATSAPI interface, defaults to the dll api. This argument
            makes it possible to provide another api, e.g. for a simulated
            driver for which the binary Alazar drivers do not need to be
            installed.
    """

    # override dll_path in your init script or in the board constructor
    # if you have it somewhere else
    dll_path = 'C:\\WINDOWS\\System32\\ATSApi'

    api: AlazarATSAPI

    # override channels in a subclass if needed
    channels = 2

    @classmethod
    def find_boards(cls, dll_path: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        Find connected Alazar boards

        Args:
            dll_path: path to the Alazar API DLL library

        Returns:
            list of board info dictionaries for each connected board
        """
        api = AlazarATSAPI(dll_path or cls.dll_path)

        system_count = api.num_of_systems()
        boards = []
        for system_id in range(1, system_count + 1):
            board_count = api.boards_in_system_by_system_id(system_id)
            for board_id in range(1, board_count + 1):
                boards.append(cls.get_board_info(api, system_id, board_id))
        return boards

    @classmethod
    def get_board_info(cls, api: AlazarATSAPI, system_id: int,
                       board_id: int) -> Dict[str, Union[str, int]]:
        """
        Get the information from a connected Alazar board

        Args:
            api: An AlazarATSAPI that wraps around the CTypes CDLL
            system_id: id of the Alazar system
            board_id: id of the board within the alazar system

        Return:

            Dictionary containing

                - system_id
                - board_id
                - board_kind (as string)
                - max_samples
                - bits_per_sample
        """
        # make a temporary instrument for this board, to make it easier
        # to get its info
        board = cls(f'alazar_temp_{time.perf_counter()}',
                    system_id=system_id, board_id=board_id)

        handle = board._handle

        board_model = api.get_board_model(handle)
        max_s, bps = api.get_channel_info_(handle)

        board.close()

        return {
            'system_id': system_id,
            'board_id': board_id,
            'board_kind': board_model,
            'max_samples': max_s,
            'bits_per_sample': bps
        }

    def __init__(
            self, name: str, system_id: int = 1, board_id: int = 1,
            dll_path: Optional[str] = None,
            api: Optional[AlazarATSAPI] = None, **kwargs: Any) -> None:
        super().__init__(name, **kwargs)
        self.api = api or AlazarATSAPI(dll_path or self.dll_path)

        self._parameters_synced = False
        self._handle = self.api.get_board_by_system_id(system_id, board_id)

        if not self._handle:
            raise Exception('AlazarTech_ATS not found at '
                            'system {}, board {}'.format(system_id, board_id))

        self.capability = CapabilityHelper(self.api, self._handle)

        self.buffer_list: List['Buffer'] = []

    def get_idn(self) -> Dict[str, Optional[Union[str, int]]]:  # type: ignore[override]
        # TODO return type is inconsistent with the super class. We should consider
        # if ints and floats are allowed as values in the dict
        """
        This methods gets the most relevant information of this instrument

        The firmware version reported should match the version number of
        downloadable fw files from AlazarTech. But note that the firmware
        version has often been found to be incorrect for several firmware
        versions. At the time of writing it is known to be correct for the
        9360 (v 21.07) and 9373 (v 30.04) but incorrect for several earlier
        versions. In Alazar DSO this is reported as FPGA Version.

        Returns:
            Dictionary containing
                - 'firmware': as string
                - 'model': as string
                - 'serial': board serial number
                - 'vendor': 'AlazarTech'
                - 'CPLD_version': version of the CPLD
                - 'driver_version': version of the driver dll
                - 'SDK_version': version of the SDK
                - 'latest_cal_date': date of the latest calibration (as string)
                - 'memory_size': size of the memory in samples
                - 'asopc_type': type of asopc (as decimal number)
                - 'pcie_link_speed': the speed of a single pcie link (in GB/s)
                - 'pcie_link_width': number of pcie links
                - 'bits_per_sample': number of bits per one sample
                - 'max_samples': board memory size in samples
        """
        max_s, bps = self.api.get_channel_info_(self._handle)
        pcie_link_speed = \
            str(self.capability.query_pcie_link_speed()) + "GB/s"

        return {
            'firmware': self.capability.query_firmware_version(),
            'model': self.api.get_board_model(self._handle),
            'max_samples': max_s,
            'bits_per_sample': bps,
            'serial': self.capability.query_serial(),
            'vendor': 'AlazarTech',
            'CPLD_version': self.api.get_cpld_version_(self._handle),
            'driver_version': self.api.get_driver_version_(),
            'SDK_version': self.api.get_sdk_version_(),
            'latest_cal_date': self.capability.query_latest_calibration(),
            'memory_size': str(self.capability.query_memory_size()),
            'asopc_type': self.capability.query_asopc_type(),
            'pcie_link_speed': pcie_link_speed,
            'pcie_link_width': str(self.capability.query_pcie_link_width())
            }

    @contextmanager
    def syncing(self) -> Iterator[None]:
        """
        Context manager for syncing settings to Alazar card. It will
        automatically call sync_settings_to_card at the end of the
        context.

        Example:
            This is intended to be used around multiple parameter sets
            to ensure syncing is done exactly once::

                with alazar.syncing():
                     alazar.trigger_source1('EXTERNAL')
                     alazar.trigger_level1(100)
        """

        yield
        self.sync_settings_to_card()

    def sync_settings_to_card(self) -> None:
        """
        Syncs all parameters to Alazar card
        """
        if self.clock_source() == 'EXTERNAL_CLOCK_10MHz_REF':
            sample_rate = self.external_sample_rate
            if self.external_sample_rate() == 'UNDEFINED':
                raise RuntimeError("Using external 10 MHz Ref but external "
                                   "sample_rate is not set")
            if self.sample_rate() != 'UNDEFINED':
                warnings.warn("Using external 10 MHz Ref but parameter sample_"
                              "rate is set. This will have no effect and "
                              "is ignored")
            # mark the unused parameter as up to date
            self.sample_rate._set_updated()
        else:
            if self.sample_rate() == 'UNDEFINED':
                raise RuntimeError(
                    "Using Internal clock but parameter sample_rate is not set")
            if self.external_sample_rate() != 'UNDEFINED':
                warnings.warn("Using Internal clock but parameter external_sample_rate is set."
                              "This will have no effect and is ignored")
            # mark the unused parameter as up to date
            self.external_sample_rate._set_updated()
            sample_rate = self.sample_rate

        self.api.set_capture_clock(
            self._handle, self.clock_source, sample_rate,
            self.clock_edge, self.decimation
        )

        for i in range(1, self.channels + 1):
            self.api.input_control(
                self._handle, 2**(i-1),
                self.parameters['coupling' + str(i)],
                self.parameters['channel_range' + str(i)],
                self.parameters['impedance' + str(i)]
            )
            if self.parameters.get('bwlimit' + str(i), None) is not None:
                self.api.set_bw_limit(
                    self._handle, 2**(i-1),
                    self.parameters['bwlimit' + str(i)]
                )

        self.api.set_trigger_operation(
            self._handle, self.trigger_operation,
            self.trigger_engine1, self.trigger_source1,
            self.trigger_slope1, self.trigger_level1,
            self.trigger_engine2, self.trigger_source2,
            self.trigger_slope2, self.trigger_level2
        )
        self.api.set_external_trigger(
            self._handle, self.external_trigger_coupling,
            self.external_trigger_range
        )
        self.api.set_trigger_delay(self._handle, self.trigger_delay)
        self.api.set_trigger_time_out(self._handle, self.timeout_ticks)
        self.api.configure_aux_io(
            self._handle, self.aux_io_mode,
            self.aux_io_param
        )
        self._parameters_synced = True

    def allocate_and_post_buffer(
            self,
            sample_type: CtypesTypes,
            n_bytes: int
    ) -> "Buffer":
        buffer = Buffer(sample_type, n_bytes)
        self.api.post_async_buffer(
            self._handle, ctypes.cast(
                buffer.addr, ctypes.c_void_p), buffer.size_bytes
        )
        return buffer

    def acquire(
            self,
            mode: Optional[str] = None,
            samples_per_record: Optional[int] = None,
            records_per_buffer: Optional[int] = None,
            buffers_per_acquisition: Optional[int] = None,
            channel_selection: Optional[str] = None,
            transfer_offset: Optional[int] = None,
            external_startcapture: Optional[str] = None,
            enable_record_headers: Optional[str] = None,
            alloc_buffers: Optional[str] = None,
            fifo_only_streaming: Optional[str] = None,
            interleave_samples:  Optional[str] = None,
            get_processed_data: Optional[str] = None,
            allocated_buffers: Optional[int] = None,
            buffer_timeout: Optional[int] = None,
            acquisition_controller: Optional["AcquisitionController[Any]"] = None
    ) -> OutputType:
        """
        perform a single acquisition with the Alazar board, and set certain
        parameters to the appropriate values
        for the parameters, see the ATS-SDK programmer's guide

        Args:
            mode:
            samples_per_record:
            records_per_buffer:
            buffers_per_acquisition:
            channel_selection:
            transfer_offset:
            external_startcapture:
            enable_record_headers:
            alloc_buffers:
            fifo_only_streaming:
            interleave_samples:
            get_processed_data:
            allocated_buffers:
            buffer_timeout:
            acquisition_controller: An instance of an acquisition controller
                that handles the dataflow of an acquisition

        Returns:
            Whatever is given by acquisition_controller.post_acquire method
        """
        if acquisition_controller is None:
            raise RuntimeError("Cannot call acquire without an "
                               "acquisition_controller")

        # region set parameters from args
        start_func = time.perf_counter()
        if self._parameters_synced is False:
            raise RuntimeError("You must sync parameters to Alazar card "
                               "before calling acquire by calling "
                               "sync_settings_to_card")
        self._set_if_present('mode', mode)
        self._set_if_present('samples_per_record', samples_per_record)
        self._set_if_present('records_per_buffer', records_per_buffer)
        self._set_if_present('buffers_per_acquisition',
                             buffers_per_acquisition)
        self._set_if_present('channel_selection', channel_selection)
        self._set_if_present('transfer_offset', transfer_offset)
        self._set_if_present('external_startcapture', external_startcapture)
        self._set_if_present('enable_record_headers', enable_record_headers)
        self._set_if_present('alloc_buffers', alloc_buffers)
        self._set_if_present('fifo_only_streaming', fifo_only_streaming)
        self._set_if_present('interleave_samples', interleave_samples)
        self._set_if_present('get_processed_data', get_processed_data)
        self._set_if_present('allocated_buffers', allocated_buffers)
        self._set_if_present('buffer_timeout', buffer_timeout)

        # endregion
        mode = self.mode.get()
        if mode not in ('TS', 'NPT'):
            raise Exception("Only the 'TS' and 'NPT' modes are implemented "
                            "at this point")

        # -----set final configurations-----

        buffers_per_acquisition = cast(int, self.buffers_per_acquisition())
        samples_per_record = cast(int, self.samples_per_record())
        records_per_buffer = cast(int, self.records_per_buffer())

        # bits per sample
        _, bits_per_sample = self.api.get_channel_info_(self._handle)

        # channels
        channels_binrep = self.channel_selection.raw_value
        number_of_channels = self.get_num_channels(channels_binrep)

        # In the following we need to consider the size of the buffer
        # in two different scenarios as several Alazar cards have sample sizes
        # that are in fractions of bytes. (such as 12 bits).
        # We are transferring data padded to
        # whole bytes. I.e a sample of 12 bits will take up 16 bits when
        # transferred so we are allocating buffers of that size.
        # However, when calculating internal limitations on the card we are
        # using the fractional sizes of samples

        # number of bytes per sample rounded up to the nearest integer
        whole_bytes_per_sample = (bits_per_sample + 7) // 8
        transfer_record_size = whole_bytes_per_sample * samples_per_record
        transfer_buffer_size = (transfer_record_size *
                                records_per_buffer * number_of_channels)

        sample_type: Union[Type[ctypes.c_uint16], Type[ctypes.c_uint8]] = (
            ctypes.c_uint16 if whole_bytes_per_sample > 1 else ctypes.c_uint8)
        internal_buffer_size_requested = (bits_per_sample * samples_per_record *
                                          records_per_buffer) // 8

        if mode == 'TS':
            transfer_buffer_size //= buffers_per_acquisition
            internal_buffer_size_requested //= buffers_per_acquisition

        if internal_buffer_size_requested > max_buffer_size:
            raise RuntimeError(f"Requested a buffer of size: "
                               f"{internal_buffer_size_requested / 1024 ** 2}"
                               f" MB. The maximum supported size is "
                               f"{max_buffer_size / 1024 ** 2} MB "
                               f"(recommended is <8MB).")

        # Set record size for NPT mode
        if mode == 'NPT':
            pretriggersize = 0  # pretriggersize is 0 for NPT always
            post_trigger_size = samples_per_record
            self.api.set_record_size(
                self._handle, pretriggersize,
                post_trigger_size
            )
        # set acquisition parameters here for NPT, TS mode
        samples_per_buffer = 0

        acquire_flags = (self.mode.raw_value |
                         self.external_startcapture.raw_value |
                         self.enable_record_headers.raw_value |
                         self.alloc_buffers.raw_value |
                         self.fifo_only_streaming.raw_value |
                         self.interleave_samples.raw_value |
                         self.get_processed_data.raw_value)

        if mode == 'NPT':
            records_per_acquisition = (
                records_per_buffer * buffers_per_acquisition)
            self.api.before_async_read(
                self._handle, self.channel_selection.raw_value,
                self.transfer_offset.raw_value,
                samples_per_record,
                records_per_buffer, records_per_acquisition,
                acquire_flags
            )
        elif mode == 'TS':
            if (samples_per_record % buffers_per_acquisition != 0):
                self.log.warning('buffers_per_acquisition is not a divisor '
                                 'of samples per record which it should be '
                                 'in TS mode, rounding down in samples per '
                                 'buffer calculation')
            samples_per_buffer = int(samples_per_record /
                                     buffers_per_acquisition)
            if self.records_per_buffer() != 1:
                self.log.warning('records_per_buffer should be 1 in TS mode, '
                                 'defauling to 1')
                self.records_per_buffer.set(1)
            records_per_buffer = cast(int, self.records_per_buffer())

            self.api.before_async_read(
                self._handle, self.channel_selection.raw_value,
                self.transfer_offset.raw_value, samples_per_buffer,
                records_per_buffer, buffers_per_acquisition,
                acquire_flags
            )

        self.clear_buffers()

        # make sure that allocated_buffers <= buffers_per_acquisition
        allocated_buffers = cast(int, self.allocated_buffers())
        buffers_per_acquisition = cast(int, self.buffers_per_acquisition())

        if allocated_buffers > buffers_per_acquisition:
            self.log.warning("'allocated_buffers' should be <= "
                             "'buffers_per_acquisition'. Defaulting "
                             "'allocated_buffers' to "
                             f"{buffers_per_acquisition}")
            self.allocated_buffers.set(buffers_per_acquisition)

        allocated_buffers = cast(int, self.allocated_buffers())
        buffer_recycling = buffers_per_acquisition > allocated_buffers

        # post buffers to Alazar
        try:
            for _ in range(allocated_buffers):
                buf = self.allocate_and_post_buffer(sample_type,
                                                    transfer_buffer_size)
                self.buffer_list.append(buf)

            # -----start capture here-----
            acquisition_controller.pre_start_capture()
            start = time.perf_counter()  # Keep track of when acquisition started
            # call the startcapture method
            self.api.start_capture(self._handle)
            acquisition_controller.pre_acquire()

            # buffer handling from acquisition
            buffers_completed = 0
            bytes_transferred = 0
            buffer_timeout = cast(int, self.buffer_timeout())

            done_setup = time.perf_counter()

            while (buffers_completed < self.buffers_per_acquisition.get()):
                # Wait for the buffer at the head of the list of available
                # buffers to be filled by the board.
                buf = self.buffer_list[buffers_completed % allocated_buffers]
                self.api.wait_async_buffer_complete(
                    self._handle,
                    ctypes.cast(buf.addr, ctypes.c_void_p),
                    buffer_timeout
                )

                acquisition_controller.buffer_done_callback(buffers_completed)

                # if buffers must be recycled, extract data and repost them
                # otherwise continue to next buffer
                if buffer_recycling:
                    acquisition_controller.handle_buffer(
                        buf.buffer, buffers_completed)
                    self.api.post_async_buffer(
                        self._handle,
                        ctypes.cast(buf.addr, ctypes.c_void_p),
                        buf.size_bytes
                    )
                buffers_completed += 1
                bytes_transferred += buf.size_bytes
        finally:
            # stop measurement here
            done_capture = time.perf_counter()
            self.api.abort_async_read(self._handle)

        time_done_abort = time.perf_counter()

        # -----cleanup here-----
        # extract data if not yet done
        if not buffer_recycling:
            for i, buf in enumerate(self.buffer_list):
                acquisition_controller.handle_buffer(buf.buffer, i)
        time_done_handling = time.perf_counter()
        # free up memory
        self.clear_buffers()

        time_done_free_mem = time.perf_counter()
        # check if all parameters are up to date
        # Getting IDN is very slow so skip that
        for _, p in self.parameters.items():
            if isinstance(p, TraceParameter):
                if p.synced_to_card == False:
                    raise RuntimeError(f"TraceParameter {p} not synced to "
                                       f"Alazar card detected. Aborting. Data "
                                       f"may be corrupt")

        # Compute the total transfer time, and display performance information.
        end_time = time.perf_counter()
        tot_time = end_time - start_func
        transfer_time_sec = end_time - start
        presetup_time = start - start_func
        setup_time = done_setup - start
        capture_time = done_capture - done_setup
        abort_time = time_done_abort - done_capture
        handling_time = time_done_handling - time_done_abort
        free_mem_time = time_done_free_mem - time_done_handling
        buffers_per_sec: float = 0
        bytes_per_sec: float = 0
        records_per_sec: float = 0
        if transfer_time_sec > 0:
            buffers_per_sec = buffers_completed / transfer_time_sec
            bytes_per_sec = bytes_transferred / transfer_time_sec
            records_per_sec = (records_per_buffer *
                               buffers_completed / transfer_time_sec)
        if self.log.isEnabledFor(logging.DEBUG):
            self.log.debug("Captured %d buffers (%f buffers per sec)" %
                           (buffers_completed, buffers_per_sec))
            self.log.debug("Captured %d records (%f records per sec)" %
                           (records_per_buffer * buffers_completed,
                            records_per_sec))
            self.log.debug("Transferred {:g} bytes ({:g} "
                           "bytes per sec)".format(bytes_transferred,
                                                   bytes_per_sec))
            self.log.debug(f"Pre setup took {presetup_time}")
            self.log.debug(f"Pre capture setup took {setup_time}")
            self.log.debug(f"Capture took {capture_time}")
            self.log.debug(f"abort took {abort_time}")
            self.log.debug(f"handling took {handling_time}")
            self.log.debug(f"free mem took {free_mem_time}")
            self.log.debug(f"tot acquire time is {tot_time}")

        # return result
        return acquisition_controller.post_acquire()

    def _set_if_present(self,
                        param_name: str,
                        value: Union[int, str, float, None]) -> None:
        if value is not None:
            parameter = self.parameters[param_name]
            parameter.set(value)

    def _set_list_if_present(self, param_base: str, value: Sequence[Union[int, str, float]]) -> None:
        if value is not None:
            for i, v in enumerate(value):
                parameter = self.parameters[param_base + str(i + 1)]
                parameter.set(v)

    def clear_buffers(self) -> None:
        """
        This method uncommits all buffers that were committed by the driver.
        This method only has to be called when the acquistion crashes, otherwise
        the driver will uncommit the buffers itself

        Returns:
            None

        """
        for b in self.buffer_list:
            b.free_mem()
        self.log.debug("buffers cleared")
        self.buffer_list = []

    def signal_to_volt(self, channel: int, signal: float) -> float:
        """
        convert a value from a buffer to an actual value in volts based on the
        ranges of the channel

        Args:
            channel: number of the channel where the signal value came from
            signal: the value that needs to be converted

        Returns:
             the corresponding value in volts
        """
        return (((signal - 127.5) / 127.5) *
                (self.parameters['channel_range' + str(channel)].get()))

    def get_sample_rate(self, include_decimation: bool = True) -> float:
        """
        Obtain the effective sampling rate of the acquisition
        based on clock speed and decimation

        Returns:
            the number of samples (per channel) per second
        """
        if (self.clock_source.get() == 'EXTERNAL_CLOCK_10MHz_REF'
                and 'external_sample_rate' in self.parameters):
            rate = self.external_sample_rate.get()
            # if we are using an external ref clock the sample rate
            # is set as an integer and not value mapped so we use a different
            # parameter to represent it
        elif self.sample_rate.get() == 'EXTERNAL_CLOCK':
            raise Exception('External clock is used, alazar driver '
                            'could not determine sample speed.')
        else:
            rate = self.sample_rate.get()
        if rate == '1GHz_REFERENCE_CLOCK':
            rate = 1e9

        if include_decimation:
            decimation = self.decimation.get()
        else:
            decimation = 0
        if decimation > 0:
            return rate / decimation
        else:
            return rate

    @staticmethod
    def get_num_channels(byte_rep: int) -> int:
        """
        Return the number of channels for a specific channel mask

        Each single channel is represented by a bitarray with one
        non zero entry i.e. powers of two. All multichannel masks can be
        constructed by summing the single channel ones. However, not all
        configurations are supported. See table 4 Input Channel Configurations
        on page 241 of the Alazar SDK manual. This contains the complete
        mapping for all current Alazar cards. It's left to the driver to
        ensure that only the ones supported for a specific card can be
        selected
        """
        n_ch = NUMBER_OF_CHANNELS_FROM_BYTE_REPR.get(byte_rep, None)
        if n_ch is None:
            raise RuntimeError(
                f'Invalid channel configuration {byte_rep!r} supplied')
        return n_ch

    def _read_register(self, offset: int) -> int:
        return self.api.read_register_(self._handle, offset)

    def _write_register(self, offset: int, value: int) -> None:
        self.api.write_register_(self._handle, offset, value)


def _setup_ctypes_for_windll_lib_functions() -> None:
    """
    Set up ``argtypes`` and ``restype`` for functions from ``ctypes.windll``
    libraries, which are used in this module.
    """
    if sys.platform == 'win32':
        ctypes.windll.kernel32.VirtualAlloc.argtypes = [
            ctypes.c_void_p,
            ctypes.c_long,
            ctypes.c_long,
            ctypes.c_long
        ]
        ctypes.windll.kernel32.VirtualAlloc.restype = ctypes.c_void_p

        ctypes.windll.kernel32.VirtualFree.argtypes = [
                ctypes.c_void_p,
                ctypes.c_long,
                ctypes.c_long
            ]
        ctypes.windll.kernel32.VirtualFree.restype = ctypes.c_int


_setup_ctypes_for_windll_lib_functions()


class Buffer:
    """Buffer suitable for DMA transfers.

    AlazarTech digitizers use direct memory access (DMA) to transfer
    data from digitizers to the computer's main memory. This class
    abstracts a memory buffer on the host, and ensures that all the
    requirements for DMA transfers are met.

    Buffer export a 'buffer' member, which is a NumPy array view
    of the underlying memory buffer

    Args:
        c_sample_type: The datatype of the buffer to create. Should be a valid
            ctypes type.
        size_bytes: The size of the buffer to allocate, in bytes.
    """

    def __init__(
            self,
            c_sample_type: CtypesTypes,
            size_bytes: int
    ):
        self.size_bytes = size_bytes
        self.buffer: np.ndarray

        npSampleType = {
            ctypes.c_uint8: np.uint8,
            ctypes.c_uint16: np.uint16,
            ctypes.c_uint32: np.uint32,
            ctypes.c_int32: np.int32,
            ctypes.c_float: np.float32
        }.get(c_sample_type, 0)

        bytes_per_sample = {
            ctypes.c_uint8:  1,
            ctypes.c_uint16: 2,
            ctypes.c_uint32: 4,
            ctypes.c_int32:  4,
            ctypes.c_float:  4
        }.get(c_sample_type, 0)

        self._allocated = True
        if sys.platform == 'win32':
            MEM_COMMIT = 0x1000
            PAGE_READWRITE = 0x4
            self.addr = ctypes.windll.kernel32.VirtualAlloc(
                0, ctypes.c_long(size_bytes), MEM_COMMIT, PAGE_READWRITE)
        else:
            self._allocated = True
            ctypes_array = (c_sample_type *
                            (size_bytes // bytes_per_sample))()
            self.addr = ctypes.addressof(ctypes_array)

        ctypes_array = (c_sample_type *
                        (size_bytes // bytes_per_sample)).from_address(self.addr)
        self.buffer = np.frombuffer(ctypes_array, dtype=npSampleType)
        self.ctypes_buffer = ctypes_array

    def free_mem(self) -> None:
        """
        uncommit memory allocated with this buffer object
        """
        self._allocated = False
        if sys.platform == 'win32':
            MEM_RELEASE = 0x8000
            ctypes.windll.kernel32.VirtualFree(
                ctypes.c_void_p(self.addr), 0, MEM_RELEASE)

    def __del__(self) -> None:
        """
        If python garbage collects this object, __del__ should be called and it
        is the last chance to uncommit the memory to prevent a memory leak.
        This method is not very reliable so users should not rely on this
        functionality
        """
        if self._allocated:
            self.free_mem()
            logger.warning(
                'Buffer prevented memory leak; Memory released to Windows.\n'
                'Memory should have been released before buffer was deleted.')


class AcquisitionInterface(Generic[OutputType]):
    """
    This class represents all choices that the end-user has to make regarding
    the data-acquisition. this class should be subclassed to program these
    choices.

    The basic structure of an acquisition is:

        - Call to :meth:`AlazarTech_ATS.acquire` internal configuration
        - Call to :meth:`AcquisitionInterface.pre_start_capture`
        - Call to the start capture of the Alazar board
        - Call to :meth:`AcquisitionInterface.pre_acquire`
        - Loop over all buffers that need to be acquired
          dump each buffer to acquisitioncontroller.handle_buffer
          (only if buffers need to be recycled to finish the acquisiton)
        - Dump remaining buffers to :meth:`AcquisitionInterface.handle_buffer`
          alazar internals
        - Return return value from :meth:`AcquisitionController.post_acquire`
    """

    def pre_start_capture(self) -> None:
        """
        Use this method to prepare yourself for the data acquisition
        The Alazar instrument will call this method right before
        'AlazarStartCapture' is called
        """
        pass

    def pre_acquire(self) -> None:
        """
        This method is called immediately after 'AlazarStartCapture' is called
        """
        pass

    def handle_buffer(self,
                      buffer: np.ndarray,
                      buffer_number: Optional[int] = None) -> None:
        """
        This method should store or process the information that is contained
        in the buffers obtained during the acquisition.

        Args:
            buffer: np.array with the data from the Alazar card
            buffer_number: counter for which buffer we are handling

        """
        raise NotImplementedError(
            'This method should be implemented in a subclass')

    def post_acquire(self) -> OutputType:
        """
        This method should return any information you want to save from this
        acquisition. The acquisition method from the Alazar driver will use
        this data as its own return value

        Returns:
            this function should return all relevant data that you want
            to get form the acquisition
        """
        raise NotImplementedError(
            'This method should be implemented in a subclass')

    def buffer_done_callback(self, buffers_completed: int) -> None:
        """
        This method is called when a buffer is completed. It can be used
        if you want to implement an event that happens for each buffer.
        You will probably want to combine this with `AUX_IN_TRIGGER_ENABLE`
        to wait before starting capture of the next buffer.

        Args:
            buffers_completed: how many buffers have been completed and copied
                to local memory at the time of this callback.
        """
        pass


class AcquisitionController(Instrument, AcquisitionInterface[Any], Generic[OutputType]):
    """
    Compatibility class. The methods of :class:`AcquisitionController`
    have been extracted. This class is the base class fro AcquisitionInterfaces
    that are intended to be QCoDeS instruments at the same time.
    """

    def __init__(self, name: str, alazar_name: str, **kwargs: Any):
        """
        Args:
            alazar_name: The name of the alazar instrument.
        """
        super().__init__(name, **kwargs)
        self._alazar: AlazarTech_ATS = cast(
            AlazarTech_ATS,
            self.find_instrument(
                alazar_name,
                instrument_class=AlazarTech_ATS)
        )

    def _get_alazar(self) -> AlazarTech_ATS:
        """
        returns a reference to the alazar instrument. A call to self._alazar is
        quicker, so use that if in need for speed
        :return: reference to the Alazar instrument
        """
        return self._alazar



import logging
import re

import pytest

from qcodes.dataset.experiment_container import (
    Experiment,
    experiments,
    load_experiment,
    load_experiment_by_name,
    load_last_experiment,
    load_or_create_experiment,
    new_experiment,
)
from qcodes.dataset.experiment_settings import (
    get_default_experiment_id,
    reset_default_experiment_id,
)
from qcodes.dataset.measurements import Measurement
from qcodes.dataset.sqlite.database import conn_from_dbpath_or_conn, get_DB_location


def assert_experiments_equal(exp, exp_2):
    for attr in ['name', 'sample_name', 'path_to_db', 'last_counter']:
        assert getattr(exp, attr) == getattr(exp_2, attr)
    assert len(exp_2) == len(exp)
    assert repr(exp_2) == repr(exp)


@pytest.mark.usefixtures("empty_temp_db")
def test_run_loaded_experiment():
    """
    Test that we can resume a measurement after loading by name
    """
    new_experiment("test", "test1")
    exp_loaded = load_experiment_by_name("test", "test1")

    meas = Measurement(exp=exp_loaded)
    meas.register_custom_parameter(name='dummy', paramtype='text')
    with meas.run():
        pass

    with meas.run():
        pass


def test_last_data_set_from_experiment(dataset):
    experiment = load_experiment(dataset.exp_id)
    ds = experiment.last_data_set()

    assert dataset.run_id == ds.run_id
    assert dataset.name == ds.name
    assert dataset.exp_id == ds.exp_id
    assert dataset.exp_name == ds.exp_name
    assert dataset.sample_name == ds.sample_name
    assert dataset.path_to_db == ds.path_to_db

    assert experiment.path_to_db == ds.path_to_db


def test_last_data_set_from_experiment_with_no_datasets(experiment):
    with pytest.raises(ValueError, match='There are no runs in this '
                                         'experiment'):
        _ = experiment.last_data_set()


@pytest.mark.usefixtures("empty_temp_db")
def test_load_or_create_experiment_loading():
    """Test that an experiment is correctly loaded"""
    exp = new_experiment("experiment_name", "sample_name")
    exp_2 = load_or_create_experiment("experiment_name", "sample_name")
    assert_experiments_equal(exp, exp_2)


@pytest.mark.usefixtures("empty_temp_db")
def test_load_or_create_experiment_different_sample_name():
    """
    Test that an experiment is created for the case when the experiment
    name is the same, but the sample name is different
    """
    exp = new_experiment("experiment_name", "sample_name_1")
    exp_2 = load_or_create_experiment("experiment_name", "sample_name_2")

    actual_experiments = experiments()
    assert len(actual_experiments) == 2

    assert exp.name == exp_2.name
    assert exp.sample_name != exp_2.sample_name


@pytest.mark.usefixtures("empty_temp_db")
def test_load_or_create_experiment_creating():
    """Test that an experiment is correctly created"""
    exp = load_or_create_experiment("experiment_name", "sample_name")
    exp_2 = load_experiment_by_name("experiment_name", "sample_name")
    assert_experiments_equal(exp, exp_2)


@pytest.mark.usefixtures("empty_temp_db")
def test_load_or_create_experiment_creating_not_empty():
    """Test that an experiment is correctly created when DB is not empty"""
    exp = load_or_create_experiment("experiment_name_1", "sample_name_1")
    exp_2 = load_or_create_experiment("experiment_name_2", "sample_name_2")

    actual_experiments = experiments()
    assert len(actual_experiments) == 2

    assert_experiments_equal(actual_experiments[0], exp)
    assert_experiments_equal(actual_experiments[1], exp_2)


@pytest.mark.usefixtures("empty_temp_db")
def test_has_attributes_after_init():
    """
    Ensure that all attributes are populated after __init__ in BOTH cases
    (exp_id is None / exp_id is not None)
    """

    attrs = ['name', 'exp_id', '_exp_id', 'sample_name', 'last_counter',
             'path_to_db', 'conn', 'started_at',
             'finished_at', 'format_string']

    # This creates an experiment in the db
    exp1 = Experiment(exp_id=None)

    # This loads the experiment that we just created
    exp2 = Experiment(exp_id=1)

    for exp in (exp1, exp2):
        for attr in attrs:
            assert hasattr(exp, attr)
            getattr(exp, attr)


def test_experiment_read_only_properties(experiment):
    read_only_props = ['name', 'exp_id', 'sample_name', 'last_counter',
                       'path_to_db', 'started_at', 'finished_at',
                       'format_string']

    # It is not expected to be possible to set read only properties
    for prop in read_only_props:
        with pytest.raises(AttributeError, match="can't set attribute"):
            setattr(experiment, prop, True)


@pytest.mark.usefixtures("empty_temp_db")
@pytest.mark.parametrize("non_existing_id", (1, 0, -1, 'number#42'))
def test_create_experiment_from_non_existing_id(non_existing_id):
    with pytest.raises(ValueError, match="No such experiment in the database"):
        _ = Experiment(exp_id=non_existing_id)


@pytest.mark.usefixtures("empty_temp_db")
@pytest.mark.parametrize("non_existing_id", (1, 0, -1))
def test_load_experiment_from_non_existing_id(non_existing_id):
    with pytest.raises(ValueError, match="No such experiment in the database"):
        _ = load_experiment(non_existing_id)


@pytest.mark.usefixtures("empty_temp_db")
@pytest.mark.parametrize("bad_id", (None, 'number#42'))
def test_load_experiment_from_bad_id(bad_id):
    with pytest.raises(ValueError, match="Experiment ID must be an integer"):
        _ = load_experiment(bad_id)


def test_format_string(empty_temp_db):
    # default format string
    exp1 = Experiment(exp_id=None)
    assert "{}-{}-{}" == exp1.format_string

    # custom format string
    fmt_str = "name_{}__id_{}__run_cnt_{}"
    exp2 = Experiment(exp_id=None, format_string=fmt_str)
    assert fmt_str == exp2.format_string

    # invalid format string
    fmt_str = "name_{}__id_{}__{}__{}"
    with pytest.raises(ValueError, match=r"Invalid format string. Can not "
                                         r"format \(name, exp_id, "
                                         r"run_counter\)"):
        _ = Experiment(exp_id=None, format_string=fmt_str)


def test_load_experiment_by_name_defaults(empty_temp_db):
    exp1 = Experiment(exp_id=None)

    exp2 = load_experiment_by_name('experiment_1')
    assert_experiments_equal(exp1, exp2)

    exp3 = load_experiment_by_name('experiment_1', 'some_sample')
    assert_experiments_equal(exp1, exp3)


def test_load_experiment_by_name(empty_temp_db):
    exp1 = Experiment(exp_id=None, name='myname')

    exp2 = load_experiment_by_name('myname')
    assert_experiments_equal(exp1, exp2)

    exp3 = load_experiment_by_name('myname', 'some_sample')
    assert_experiments_equal(exp1, exp3)

    exp4 = Experiment(exp_id=None, name='with_sample', sample_name='mysample')

    exp5 = load_experiment_by_name('with_sample')
    assert_experiments_equal(exp4, exp5)

    exp6 = load_experiment_by_name('with_sample', 'mysample')
    assert_experiments_equal(exp4, exp6)


def test_load_experiment_by_name_bad_name(empty_temp_db):
    Experiment(exp_id=None, name='myname')

    with pytest.raises(ValueError, match='Experiment not found'):
        _ = load_experiment_by_name('myname__')

    with pytest.raises(ValueError, match='Experiment not found'):
        _ = load_experiment_by_name('myname__', 'some_sample')

    with pytest.raises(ValueError, match='Experiment not found'):
        _ = load_experiment_by_name('myname__', 'some_sample__')


def test_load_experiment_by_name_bad_sample_name(empty_temp_db):
    Experiment(exp_id=None, sample_name='mysample')

    with pytest.raises(ValueError, match='Experiment not found'):
        _ = load_experiment_by_name('experiment_1', 'mysample__')

    with pytest.raises(ValueError, match='Experiment not found'):
        _ = load_experiment_by_name('experiment_1__', 'mysample')

    with pytest.raises(ValueError, match='Experiment not found'):
        _ = load_experiment_by_name('experiment_1__', 'mysample__')


def test_load_experiment_by_name_duplicate_name(empty_temp_db):
    exp1 = Experiment(exp_id=None, name="exp")
    exp2 = Experiment(exp_id=None, name="exp")

    repr_str_1_2 = (
        f"Many experiments matching your request found:\n"
        f"exp_id:{exp1.exp_id} ({exp1.name}-{exp1.sample_name}) "
        f"started at ({exp1.started_at})\n"
        f"exp_id:{exp2.exp_id} ({exp2.name}-{exp2.sample_name}) "
        f"started at ({exp2.started_at})"
    )
    repr_str_1_2_regex = re.escape(repr_str_1_2)
    with pytest.raises(ValueError, match=repr_str_1_2_regex):
        load_experiment_by_name("exp", "some_sample")

    last_exp = load_experiment_by_name("exp", load_last_duplicate=True)
    assert last_exp.name == "exp"
    assert last_exp.sample_name == "some_sample"
    assert last_exp.exp_id == 2

    exp3 = Experiment(exp_id=None, name="exp", sample_name="my_sample")
    repr_str_1_2_3 = (
        repr_str_1_2 + "\n" + f"exp_id:{exp3.exp_id} ({exp3.name}-{exp3.sample_name}) "
        f"started at ({exp3.started_at})"
    )
    repr_str_1_2_3_regex = re.escape(repr_str_1_2_3)
    with pytest.raises(ValueError, match=repr_str_1_2_3_regex):
        load_experiment_by_name("exp")
    last_exp = load_experiment_by_name("exp", load_last_duplicate=True)
    assert last_exp.name == "exp"
    assert last_exp.sample_name == "my_sample"
    assert last_exp.exp_id == 3

    exp3_loaded = load_experiment_by_name('exp', 'my_sample')
    assert_experiments_equal(exp3, exp3_loaded)


def test_load_experiment_by_name_duplicate_sample_name(empty_temp_db):
    exp1 = Experiment(exp_id=None, name='exp1', sample_name='sss')
    exp2 = Experiment(exp_id=None, name='exp2', sample_name='sss')

    exp1_loaded = load_experiment_by_name('exp1')
    assert_experiments_equal(exp1, exp1_loaded)

    exp2_loaded = load_experiment_by_name('exp2')
    assert_experiments_equal(exp2, exp2_loaded)

    exp1_loaded_with_sample = load_experiment_by_name('exp1', 'sss')
    assert_experiments_equal(exp1, exp1_loaded_with_sample)

    exp2_loaded_with_sample = load_experiment_by_name('exp2', 'sss')
    assert_experiments_equal(exp2, exp2_loaded_with_sample)


def test_load_experiment_by_name_duplicate_name_and_sample_name(empty_temp_db):
    exp1 = Experiment(exp_id=None, name='exp', sample_name='sss')
    exp2 = Experiment(exp_id=None, name='exp', sample_name='sss')

    repr_str = f"Many experiments matching your request found:\n" \
               f"exp_id:{exp1.exp_id} ({exp1.name}-{exp1.sample_name}) " \
               f"started at ({exp1.started_at})\n" \
               f"exp_id:{exp2.exp_id} ({exp2.name}-{exp2.sample_name}) " \
               f"started at ({exp2.started_at})"
    repr_str_regex = re.escape(repr_str)

    with pytest.raises(ValueError, match=repr_str_regex):
        load_experiment_by_name('exp')

    with pytest.raises(ValueError, match=repr_str_regex):
        load_experiment_by_name("exp", "sss")

    last_exp = load_experiment_by_name("exp", "sss", load_last_duplicate=True)
    assert last_exp.name == "exp"
    assert last_exp.sample_name == "sss"
    assert last_exp.exp_id == 2


def test_new_experiment_duplicate_name_and_sample_name(empty_temp_db, caplog):
    """
    Test new_experiment to raise warning if it wants to create experiment
    with a duplicate experiment name and sample name.
    """
    exp_1 = new_experiment("exp", "sample")
    warn_msg = (
        f"There is (are) already experiment(s) with the name of {exp_1.name} "
        f"and sample name of {exp_1.sample_name} in the database."
    )
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        new_experiment("exp", "sample")
        for record in caplog.records:
            assert record.levelname == "WARNING"
        assert warn_msg in caplog.text

    exp_2 = new_experiment("exp_2", None)
    warn_msg = (
        f"There is (are) already experiment(s) with the name of {exp_2.name} "
        f"and sample name of {exp_2.sample_name} in the database."
    )
    assert exp_2.sample_name == "some_sample"
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        new_experiment("exp_2", None)
        for record in caplog.records:
            assert record.levelname == "WARNING"
        assert warn_msg in caplog.text
    caplog.clear()


def test_load_last_experiment(empty_temp_db):
    # test in case of no experiments
    with pytest.raises(ValueError, match='There are no experiments in the '
                                         'database file'):
        _ = load_last_experiment()

    # create 2 experiments
    exp1 = Experiment(exp_id=None)
    exp2 = Experiment(exp_id=None)
    assert get_DB_location() == exp1.path_to_db
    assert get_DB_location() == exp2.path_to_db

    # load last and assert that its the 2nd one that was created
    last_exp = load_last_experiment()
    assert last_exp.exp_id == exp2.exp_id
    assert last_exp.exp_id != exp1.exp_id
    assert last_exp.path_to_db == exp2.path_to_db


def test_active_experiment(empty_temp_db):

    conn = conn_from_dbpath_or_conn(conn=None, path_to_db=empty_temp_db)
    with pytest.raises(ValueError):
        get_default_experiment_id(conn)

    exp_1 = load_or_create_experiment("test_exp", sample_name="no_sample")
    assert get_default_experiment_id(conn) == exp_1.exp_id

    exp_2 = new_experiment("test_exp_2", sample_name="no_sample")
    assert get_default_experiment_id(conn) == exp_2.exp_id

    exp_3 = load_experiment(1)
    assert get_default_experiment_id(conn) == exp_1.exp_id
    assert get_default_experiment_id(conn) == exp_3.exp_id

    exp_4 = new_experiment("test_exp_3", sample_name="no_sample")

    exp_5 = load_experiment_by_name("test_exp_2", sample="no_sample")
    assert get_default_experiment_id(conn) == exp_2.exp_id
    assert get_default_experiment_id(conn) == exp_5.exp_id

    exp_6 = load_last_experiment()
    assert get_default_experiment_id(conn) == exp_4.exp_id
    assert get_default_experiment_id(conn) == exp_6.exp_id

    last_exp = new_experiment("last_exp", sample_name="no_sample")
    load_experiment(3)

    reset_default_experiment_id(conn)
    assert get_default_experiment_id(conn) is last_exp.exp_id

    load_experiment(exp_1.exp_id)
    assert get_default_experiment_id(conn) == exp_1.exp_id

    reset_default_experiment_id()
    assert get_default_experiment_id(conn) is last_exp.exp_id



import pytest
from qcodes.utils.validators import Ints, Sequence


def test_type():
    l = Sequence()
    v1 = ['a', 'b', 5]
    l.validate(v1)

    l.validate((1, 2, 3))

    v2 = 234
    with pytest.raises(TypeError):
        l.validate(v2)


def test_elt_vals():
    l = Sequence(Ints(max_value=10))
    v1 = [0, 1, 5]
    l.validate(v1)

    v2 = [0, 1, 11]
    with pytest.raises(ValueError, match="11 is invalid: must be between"):
        l.validate(v2)


def test_valid_values():
    val = Sequence(Ints(max_value=10))
    for vval in val.valid_values:
        val.validate(vval)


def test_length():
    l = Sequence(length=3)
    v1 = [0, 1, 5]
    l.validate(v1)

    v2 = [0, 1, 3, 4]
    with pytest.raises(ValueError, match="has not length"):
        l.validate(v2)

    v3 = [3, 4]
    with pytest.raises(ValueError, match="has not length"):
        l.validate(v3)


def test_sorted():
    l = Sequence(length=3, require_sorted=True)

    v1 = [0, 1, 5]
    l.validate(v1)

    v2 = (1, 3, 5)
    l.validate(v2)

    v3 = (1, 5, 3)
    with pytest.raises(ValueError, match="is required to be sorted"):
        l.validate(v3)

    v4 = [1, 7, 2]
    with pytest.raises(ValueError, match="is required to be sorted"):
        l.validate(v4)



"""
This module provides means of connecting to a QCoDeS database file and
initialising it. Note that connecting/initialisation take into account
database version and possibly perform database upgrades.
"""
import io
import sqlite3
import sys
from contextlib import contextmanager
from os.path import expanduser, normpath
from pathlib import Path
from typing import Iterator, Optional, Tuple, Union

import numpy as np

import qcodes
from qcodes.dataset.experiment_settings import reset_default_experiment_id
from qcodes.dataset.sqlite.connection import ConnectionPlus
from qcodes.dataset.sqlite.db_upgrades import (
    _latest_available_version,
    get_user_version,
    perform_db_upgrade,
)
from qcodes.dataset.sqlite.initial_schema import init_db
from qcodes.utils.types import (
    complex_type_union,
    complex_types,
    numpy_floats,
    numpy_ints,
)


# utility function to allow sqlite/numpy type
def _adapt_array(arr: np.ndarray) -> sqlite3.Binary:
    """
    See this:
    https://stackoverflow.com/questions/3425320/sqlite3-programmingerror-you-must-not-use-8-bit-bytestrings-unless-you-use-a-te
    """
    out = io.BytesIO()
    np.save(out, arr)
    out.seek(0)
    return sqlite3.Binary(out.read())


def _convert_array(text: bytes) -> np.ndarray:
    out = io.BytesIO(text)
    out.seek(0)
    return np.load(out)


def _convert_complex(text: bytes) -> complex_type_union:
    out = io.BytesIO(text)
    out.seek(0)
    return np.load(out)[0]


this_session_default_encoding = sys.getdefaultencoding()


def _convert_numeric(value: bytes) -> Union[float, int, str]:
    """
    This is a converter for sqlite3 'numeric' type class.

    This converter is capable of deducting whether a number is a float or an
    int.

    Note sqlite3 allows to save data to columns even if their type is not
    compatible with the table type class (for example, it is possible to save
    integers into 'text' columns). Due to this fact, and for the reasons of
    flexibility, the numeric converter is also made capable of handling
    strings. An obvious exception to this is 'nan' (case insensitive) which
    gets converted to `np.nan`. Another exception to this is 'inf', which
    gets converted to 'np.inf'.
    """
    try:
        # First, try to convert bytes to float
        numeric = float(value)
    except ValueError as e:
        # If an exception has been raised, we first need to find out
        # if the reason was the conversion to float, and, if so, we are sure
        # that we need to return a string
        if "could not convert string to float" in str(e):
            return str(value, encoding=this_session_default_encoding)
        else:
            # otherwise, the exception is forwarded up the stack
            raise e

    # If that worked, e.g. did not raise an exception, then we check if the
    # outcome is 'nan'
    if np.isnan(numeric):
        return numeric

    # Then we check if the outcome is 'inf', includes +inf and -inf
    if np.isinf(numeric):
        return numeric

    # If it is not 'nan' and not 'inf', then we need to see if the value is
    # really an integer or with floating point digits
    numeric_int = int(numeric)
    if numeric != numeric_int:
        return numeric
    else:
        return numeric_int


def _adapt_float(fl: float) -> Union[float, str]:
    if np.isnan(fl):
        return "nan"
    return float(fl)


def _adapt_complex(value: complex_type_union) -> sqlite3.Binary:
    out = io.BytesIO()
    np.save(out, np.array([value]))
    out.seek(0)
    return sqlite3.Binary(out.read())


def connect(name: Union[str, Path], debug: bool = False,
            version: int = -1) -> ConnectionPlus:
    """
    Connect or create  database. If debug the queries will be echoed back.
    This function takes care of registering the numpy/sqlite type
    converters that we need.

    Args:
        name: name or path to the sqlite file
        debug: whether or not to turn on tracing
        version: which version to create. We count from 0. -1 means 'latest'.
            Should always be left at -1 except when testing.

    Returns:
        conn: connection object to the database (note, it is
            `ConnectionPlus`, not `sqlite3.Connection`

    """
    # register numpy->binary(TEXT) adapter
    sqlite3.register_adapter(np.ndarray, _adapt_array)
    # register binary(TEXT) -> numpy converter
    sqlite3.register_converter("array", _convert_array)

    sqlite3_conn = sqlite3.connect(name, detect_types=sqlite3.PARSE_DECLTYPES,
                                   check_same_thread=True)
    conn = ConnectionPlus(sqlite3_conn)

    latest_supported_version = _latest_available_version()
    db_version = get_user_version(conn)

    if db_version > latest_supported_version:
        raise RuntimeError(f"Database {name} is version {db_version} but this "
                           f"version of QCoDeS supports up to "
                           f"version {latest_supported_version}")

    # sqlite3 options
    conn.row_factory = sqlite3.Row

    # Make sure numpy ints and floats types are inserted properly
    for numpy_int in numpy_ints:
        sqlite3.register_adapter(numpy_int, int)

    sqlite3.register_converter("numeric", _convert_numeric)

    for numpy_float in (float,) + numpy_floats:
        sqlite3.register_adapter(numpy_float, _adapt_float)

    for complex_type in complex_types:
        sqlite3.register_adapter(complex_type, _adapt_complex)
    sqlite3.register_converter("complex", _convert_complex)

    if debug:
        conn.set_trace_callback(print)

    init_db(conn)
    perform_db_upgrade(conn, version=version)
    return conn


def get_db_version_and_newest_available_version(path_to_db: str) -> Tuple[int,
                                                                          int]:
    """
    Connect to a DB without performing any upgrades and get the version of
    that database file along with the newest available version (the one that
    a normal "connect" will automatically upgrade to)

    Args:
        path_to_db: the absolute path to the DB file

    Returns:
        A tuple of (db_version, latest_available_version)
    """
    conn = connect(path_to_db, version=0)
    db_version = get_user_version(conn)

    return db_version, _latest_available_version()


def get_DB_location() -> str:
    return normpath(expanduser(qcodes.config["core"]["db_location"]))


def get_DB_debug() -> bool:
    return bool(qcodes.config["core"]["db_debug"])


def initialise_database(journal_mode: Optional[str] = 'WAL') -> None:
    """
    Initialise a database in the location specified by the config object
    and set ``atomic commit and rollback mode`` of the db. The db is created
    with the latest supported version. If the database already exists the
    ``atomic commit and rollback mode`` is set and the database is upgraded
    to the latest version.

    Args:
        journal_mode: Which `journal_mode` should be used for atomic commit and rollback.
            Options are DELETE, TRUNCATE, PERSIST, MEMORY, WAL and OFF. If set to None
            no changes are made.
    """
    # calling connect performs all the needed actions to create and upgrade
    # the db to the latest version.
    conn = connect(get_DB_location(), get_DB_debug())
    reset_default_experiment_id(conn)
    if journal_mode is not None:
        set_journal_mode(conn, journal_mode)
    conn.close()
    del conn


def set_journal_mode(conn: ConnectionPlus, journal_mode: str) -> None:
    """
    Set the ``atomic commit and rollback mode`` of the sqlite database.
    See https://www.sqlite.org/pragma.html#pragma_journal_mode for details.

    Args:
        conn: Connection to the database.
        journal_mode: Which `journal_mode` should be used for atomic commit and rollback.
            Options are DELETE, TRUNCATE, PERSIST, MEMORY, WAL and OFF. If set to None
            no changes are made.
    """
    valid_journal_modes = ["DELETE", "TRUNCATE", "PERSIST", "MEMORY", "WAL", "OFF"]
    if journal_mode not in valid_journal_modes:
        raise RuntimeError(f"Invalid journal_mode {journal_mode} "
                           f"Valid modes are {valid_journal_modes}")
    query = f"PRAGMA journal_mode={journal_mode};"
    cursor = conn.cursor()
    cursor.execute(query)


def initialise_or_create_database_at(db_file_with_abs_path: str,
                                     journal_mode: Optional[str] = 'WAL') -> None:
    """
    This function sets up QCoDeS to refer to the given database file. If the
    database file does not exist, it will be initiated.

    Args:
        db_file_with_abs_path
            Database file name with absolute path, for example
            ``C:\\mydata\\majorana_experiments.db``
        journal_mode: Which `journal_mode` should be used for atomic commit and rollback.
            Options are DELETE, TRUNCATE, PERSIST, MEMORY, WAL and OFF. If set to None
            no changes are made.
    """
    qcodes.config.core.db_location = db_file_with_abs_path
    initialise_database(journal_mode)


@contextmanager
def initialised_database_at(db_file_with_abs_path: str) -> Iterator[None]:
    """
    Initializes or creates a database and restores the 'db_location' afterwards.

    Args:
        db_file_with_abs_path
            Database file name with absolute path, for example
            ``C:\\mydata\\majorana_experiments.db``
    """
    db_location = qcodes.config["core"]["db_location"]
    try:
        initialise_or_create_database_at(db_file_with_abs_path)
        yield
    finally:
        qcodes.config["core"]["db_location"] = db_location


def conn_from_dbpath_or_conn(conn: Optional[ConnectionPlus],
                             path_to_db: Optional[str]) \
        -> ConnectionPlus:
    """
    A small helper function to abstract the logic needed for functions
    that take either a `ConnectionPlus` or the path to a db file.
    If neither is given this will fall back to the default db location.
    It is an error to supply both.

    Args:
        conn: A ConnectionPlus object pointing to a sqlite database
        path_to_db: The path to a db file.

    Returns:
        A `ConnectionPlus` object
    """

    if path_to_db is not None and conn is not None:
        raise ValueError('Received BOTH conn and path_to_db. Please '
                         'provide only one or the other.')
    if conn is None and path_to_db is None:
        path_to_db = get_DB_location()

    if conn is None and path_to_db is not None:
        conn = connect(path_to_db, get_DB_debug())
    elif conn is not None:
        conn = conn
    else:
        # this should be impossible but left here to keep mypy happy.
        raise RuntimeError("Could not obtain a connection from"
                           "supplied information.")
    return conn



import os
from warnings import warn

import numpy as np

from qcodes.dataset.data_set import DataSet
from qcodes.dataset.descriptions.versioning.converters import new_to_old
from qcodes.dataset.experiment_container import _create_exp_if_needed
from qcodes.dataset.linked_datasets.links import links_to_str
from qcodes.dataset.sqlite.connection import ConnectionPlus, atomic
from qcodes.dataset.sqlite.database import (
    connect,
    get_db_version_and_newest_available_version,
)
from qcodes.dataset.sqlite.queries import (
    _populate_results_table,
    _rewrite_timestamps,
    add_data_to_dynamic_columns,
    create_run,
    get_exp_ids_from_run_ids,
    get_experiment_attributes_by_exp_id,
    get_runid_from_guid,
    is_run_id_in_database,
    mark_run_complete,
)


def extract_runs_into_db(source_db_path: str,
                         target_db_path: str, *run_ids: int,
                         upgrade_source_db: bool = False,
                         upgrade_target_db: bool = False) -> None:
    """
    Extract a selection of runs into another DB file. All runs must come from
    the same experiment. They will be added to an experiment with the same name
    and ``sample_name`` in the target db. If such an experiment does not exist, it
    will be created.

    Args:
        source_db_path: Path to the source DB file
        target_db_path: Path to the target DB file. The target DB file will be
          created if it does not exist.
        run_ids: The ``run_id``'s of the runs to copy into the target DB file
        upgrade_source_db: If the source DB is found to be in a version that is
          not the newest, should it be upgraded?
        upgrade_target_db: If the target DB is found to be in a version that is
          not the newest, should it be upgraded?
    """
    # Check for versions
    (s_v, new_v) = get_db_version_and_newest_available_version(source_db_path)
    if s_v < new_v and not upgrade_source_db:
        warn(f'Source DB version is {s_v}, but this function needs it to be'
             f' in version {new_v}. Run this function again with '
             'upgrade_source_db=True to auto-upgrade the source DB file.')
        return

    if os.path.exists(target_db_path):
        (t_v, new_v) = get_db_version_and_newest_available_version(target_db_path)
        if t_v < new_v and not upgrade_target_db:
            warn(f'Target DB version is {t_v}, but this function needs it to '
                 f'be in version {new_v}. Run this function again with '
                 'upgrade_target_db=True to auto-upgrade the target DB file.')
            return

    source_conn = connect(source_db_path)

    # Validate that all runs are in the source database
    do_runs_exist = is_run_id_in_database(source_conn, *run_ids)
    if False in do_runs_exist.values():
        source_conn.close()
        non_existing_ids = [rid for rid in run_ids if not do_runs_exist[rid]]
        err_mssg = ("Error: not all run_ids exist in the source database. "
                    "The following run(s) is/are not present: "
                    f"{non_existing_ids}")
        raise ValueError(err_mssg)

    # Validate that all runs are from the same experiment

    source_exp_ids = np.unique(get_exp_ids_from_run_ids(source_conn, run_ids))
    if len(source_exp_ids) != 1:
        source_conn.close()
        raise ValueError('Did not receive runs from a single experiment. '
                         f'Got runs from experiments {source_exp_ids}')

    # Fetch the attributes of the runs' experiment
    # hopefully, this is enough to uniquely identify the experiment
    exp_attrs = get_experiment_attributes_by_exp_id(source_conn, source_exp_ids[0])

    # Massage the target DB file to accommodate the runs
    # (create new experiment if needed)

    target_conn = connect(target_db_path)

    # this function raises if the target DB file has several experiments
    # matching both the name and sample_name

    try:
        with atomic(target_conn) as target_conn:

            target_exp_id = _create_exp_if_needed(target_conn,
                                                  exp_attrs['name'],
                                                  exp_attrs['sample_name'],
                                                  exp_attrs['format_string'],
                                                  exp_attrs['start_time'],
                                                  exp_attrs['end_time'])

            # Finally insert the runs
            for run_id in run_ids:
                _extract_single_dataset_into_db(DataSet(run_id=run_id,
                                                        conn=source_conn),
                                                target_conn,
                                                target_exp_id)
    finally:
        source_conn.close()
        target_conn.close()


def _extract_single_dataset_into_db(dataset: DataSet,
                                    target_conn: ConnectionPlus,
                                    target_exp_id: int) -> None:
    """
    NB: This function should only be called from within
    meth:`extract_runs_into_db`

    Insert the given dataset into the specified database file as the latest
    run.

    Trying to insert a run already in the DB is a NOOP.

    Args:
        dataset: A dataset representing the run to be copied
        target_conn: connection to the DB. Must be atomically guarded
        target_exp_id: The ``exp_id`` of the (target DB) experiment in which to
          insert the run
    """

    if not dataset.completed:
        raise ValueError('Dataset not completed. An incomplete dataset '
                         'can not be copied. The incomplete dataset has '
                         f'GUID: {dataset.guid} and run_id: {dataset.run_id}')

    source_conn = dataset.conn

    run_id = get_runid_from_guid(target_conn, dataset.guid)

    if run_id is not None:
        return

    target_table_name = _add_run_to_runs_table(dataset, target_conn, target_exp_id)

    _populate_results_table(
        source_conn, target_conn, dataset.table_name, target_table_name
    )


def _add_run_to_runs_table(
    dataset: DataSet, target_conn: ConnectionPlus, target_exp_id: int
) -> str:
    if dataset.parameters is not None:
        param_names = dataset.parameters.split(',')
    else:
        param_names = []
    parspecs_dict = {
        p.name: p for p in new_to_old(dataset.description.interdeps).paramspecs
    }
    parspecs = [parspecs_dict[p] for p in param_names]
    metadata = dataset.metadata
    snapshot_raw = dataset.snapshot_raw
    captured_run_id = dataset.captured_run_id
    captured_counter = dataset.captured_counter
    parent_dataset_links = links_to_str(dataset.parent_dataset_links)
    _, target_run_id, target_table_name = create_run(
        target_conn,
        target_exp_id,
        name=dataset.name,
        guid=dataset.guid,
        parameters=parspecs,
        metadata=metadata,
        captured_run_id=captured_run_id,
        captured_counter=captured_counter,
        parent_dataset_links=parent_dataset_links,
    )
    assert target_table_name is not None
    mark_run_complete(target_conn, target_run_id)
    _rewrite_timestamps(target_conn,
                        target_run_id,
                        dataset.run_timestamp_raw,
                        dataset.completed_timestamp_raw)
    if snapshot_raw is not None:
        add_data_to_dynamic_columns(
            target_conn, target_run_id, {"snapshot": snapshot_raw}
        )
    return target_table_name



"""
This module provides an api that can be passed instead for the
:class:`qcodes.instrument_drivers.AlazarTech.ats_api.AlazarATSAPI`
, which does not require a real alazar card present in the system.
This mock does currently represent a ATS9360 and does not provide
any functionality whatsoever.
"""


from typing import Tuple, Callable, Optional, Dict, Any
import numpy as np
import ctypes

from qcodes.instrument_drivers.AlazarTech.dll_wrapper import (
    _mark_params_as_updated, ReturnCode)
from qcodes.instrument_drivers.AlazarTech.constants import (
    Capability, API_SUCCESS)
from qcodes.instrument_drivers.AlazarTech.ats_api import AlazarATSAPI


class SimulatedATS9360API(AlazarATSAPI):

    registers = {
        8: 70254688,
        58: int(np.uint32(1 << 26))  # Trigger hold off
    }

    def __init__(
            self,
            dll_path: Optional[str] = None,  # Need this for meta super class
            buffer_generator: Optional[Callable[[np.ndarray], None]] = None,
            dtype: type = np.uint16
    ):
        def _default_buffer_generator(buffer: np.ndarray) -> None:
            upper = buffer.size // 2
            buffer[0:upper] = 30000 * np.ones(upper)

        # don't call `super().__init__` here to avoid dependence on the
        # alazar driver, when loading the dll
        self._buffer_generator = (
            buffer_generator or _default_buffer_generator)
        self.dtype = dtype
        self.buffers: Dict[int, np.ndarray] = {}

    def _sync_dll_call(self, c_name: str, *args: Any) -> None:
        _mark_params_as_updated(*args)

    def get_board_by_system_id(self, system_id: int, board_id: int) -> int:
        # This is the handle....
        return 12335

    def busy(self, handle: int) -> int:
        return 0

    # OTHER API-RELATED METHODS

    def get_board_model(self, handle: int) -> str:
        return 'ATS9360'

    def get_channel_info_(self, handle: int) -> Tuple[int, int]:
        return 4294967294, 12

    def get_cpld_version_(self, handle: int) -> str:
        return "25.16"

    def get_driver_version_(self) -> str:
        return '6.5.1'

    def get_sdk_version_(self) -> str:
        return '6.5.1'

    def query_capability_(self, handle: int, capability: int) -> int:
        capabilities = {
            Capability.ASOPC_TYPE: 1763017568,
            Capability.GET_SERIAL_NUMBER: 970396,
            Capability.MEMORY_SIZE: 4294967294,
            Capability.GET_PCIE_LINK_WIDTH: 8,
            Capability.GET_PCIE_LINK_SPEED: 2,
            Capability.GET_LATEST_CAL_DATE: 250117}
        return capabilities[Capability(capability)]

    def read_register_(self, handle: int, offset: int) -> int:
        return self.registers[offset]

    def write_register_(self, handle: int, offset: int, value: int) -> None:
        self.registers[offset] = value

    def post_async_buffer(
            self,
            handle: int,
            buffer: ctypes.c_void_p,
            buffer_length: int
    ) -> ReturnCode:
        if buffer.value is None:
            raise RuntimeError(
                '`post_async_buffer` received buffer with invalid address.')
        ctypes_array = (ctypes.c_uint16 *
                        (buffer_length // 2)).from_address(buffer.value)
        self.buffers[buffer.value] = np.frombuffer(
            ctypes_array, dtype=self.dtype)
        self._sync_dll_call(
            'AlazarPostAsyncBuffer', handle, buffer, buffer_length)
        return API_SUCCESS

    def wait_async_buffer_complete(
            self,
            handle: int,
            buffer: ctypes.c_void_p,
            timeout_in_ms: int
    ) -> ReturnCode:
        if buffer.value is None:
            raise RuntimeError(
                '`wait_async_buffer` received buffer with invalid address.')
        b = self.buffers.get(buffer.value)
        if b is None:
            raise RuntimeError("received an empty buffer")
        self._buffer_generator(b)
        self._sync_dll_call(
            'AlazarWaitAsyncBufferComplete', handle, buffer, timeout_in_ms)
        return API_SUCCESS



import pytest

from qcodes.instrument.parameter import invert_val_mapping
from qcodes.utils.helpers import create_on_off_val_mapping


def test_values_of_mapping_are_only_the_given_two():
    val_mapping = create_on_off_val_mapping(on_val='666', off_val='000')
    values_set = set(list(val_mapping.values()))
    assert values_set == {'000', '666'}


def test_its_inverse_maps_only_to_booleans():
    inverse = invert_val_mapping(
        create_on_off_val_mapping(on_val='666', off_val='000'))

    assert inverse == {'666': True, '000': False}


@pytest.mark.parametrize(('on_val', 'off_val'),
                         ((1, 0),
                          (1.0, 0.0),
                          ('1', '0'),
                          (True, False)
                         ))
def test_create_on_off_val_mapping_for(on_val, off_val):
    """
    Explicitly test ``create_on_off_val_mapping`` function
    by covering some of the edge cases of ``on_val`` and ``off_val``
    """
    val_mapping = create_on_off_val_mapping(on_val=on_val,
                                            off_val=off_val)

    values_list = list(set(val_mapping.values()))

    assert len(values_list) == 2
    assert on_val in values_list
    assert off_val in values_list

    assert val_mapping[1] is on_val
    assert val_mapping[True] is on_val
    assert val_mapping['1'] is on_val
    assert val_mapping['ON'] is on_val
    assert val_mapping['On'] is on_val
    assert val_mapping['on'] is on_val

    assert val_mapping[0] is off_val
    assert val_mapping[False] is off_val
    assert val_mapping['0'] is off_val
    assert val_mapping['OFF'] is off_val
    assert val_mapping['Off'] is off_val
    assert val_mapping['off'] is off_val

    inverse = invert_val_mapping(val_mapping)

    assert inverse[on_val] is True
    assert inverse[off_val] is False



from distutils.version import LooseVersion
from typing import Any

import numpy as np

from qcodes.instrument_drivers.AlazarTech.ATS import AlazarTech_ATS
from qcodes.instrument_drivers.AlazarTech.utils import TraceParameter
from qcodes.utils import validators


class AlazarTech_ATS9373(AlazarTech_ATS):
    """
    This class is the driver for the ATS9373 board.

    Note that this board is very similar to ATS9360. Refer to
    ATS SDK for details.

    Note that channels of this board have 12-bit resolution
    (see `IDN()['bits_per_sample']`) which means that the
    raw data that is returned by the card should be converted to
    uint16 type with a bit shift by 4 bits. Refer to ATS SDK
    for more infromation.
    """
    samples_divisor = 128
    _trigger_holdoff_min_fw_version = '30.04'

    def __init__(self, name: str,
                 dll_path: str = 'C:\\WINDOWS\\System32\\ATSApi.dll',
                 **kwargs: Any):
        super().__init__(name, dll_path=dll_path, **kwargs)

        # add parameters

        # ----- Parameters for the configuration of the board -----
        self.add_parameter(name='clock_source',
                           parameter_class=TraceParameter,
                           label='Clock Source',
                           unit=None,
                           initial_value='INTERNAL_CLOCK',
                           val_mapping={'INTERNAL_CLOCK': 1,
                                        'FAST_EXTERNAL_CLOCK': 2,
                                        'EXTERNAL_CLOCK_10MHz_REF': 7})
        self.add_parameter(name='external_sample_rate',
                           parameter_class=TraceParameter,
                           label='External Sample Rate',
                           unit='S/s',
                           vals=validators.MultiType(validators.Ints(300000000, 2000000000),
                                                     validators.Enum('UNDEFINED')),
                           initial_value='UNDEFINED')
        self.add_parameter(name='sample_rate',
                           parameter_class=TraceParameter,
                           label='Internal Sample Rate',
                           unit='S/s',
                           initial_value='UNDEFINED',
                           val_mapping={1_000: 1,
                                        2_000: 2,
                                        5_000: 4,
                                       10_000: 8,
                                       20_000: 10,
                                       50_000: 12,
                                      100_000: 14,
                                      200_000: 16,
                                      500_000: 18,
                                    1_000_000: 20,
                                    2_000_000: 24,
                                    5_000_000: 26,
                                   10_000_000: 28,
                                   20_000_000: 30,
                                   25_000_000: 33,
                                   50_000_000: 34,
                                  100_000_000: 36,
                                  125_000_000: 37,
                                  160_000_000: 38,
                                  180_000_000: 39,
                                  200_000_000: 40,
                                  250_000_000: 43,
                                  500_000_000: 48,
                                  800_000_000: 50,
                                1_000_000_000: 53,
                                1_200_000_000: 55,
                                1_500_000_000: 58,
                                1_800_000_000: 61,
                                2_000_000_000: 63,
                                2_400_000_000: 106,
                                3_000_000_000: 117,
                                3_600_000_000: 123,
                                4_000_000_000: 128,
                             'EXTERNAL_CLOCK': 64,
                                  'UNDEFINED': 'UNDEFINED'})
        self.add_parameter(name='clock_edge',
                           parameter_class=TraceParameter,
                           label='Clock Edge',
                           unit=None,
                           initial_value='CLOCK_EDGE_RISING',
                           val_mapping={'CLOCK_EDGE_RISING': 0,
                                        'CLOCK_EDGE_FALLING': 1})
        self.add_parameter(name='decimation',
                           parameter_class=TraceParameter,
                           label='Decimation',
                           unit=None,
                           initial_value=1,
                           vals=validators.Ints(0, 100000))

        for i in range(1, self.channels+1):
            self.add_parameter(name=f'coupling{i}',
                               parameter_class=TraceParameter,
                               label=f'Coupling channel {i}',
                               unit=None,
                               initial_value='DC',
                               val_mapping={'AC': 1, 'DC': 2})
            self.add_parameter(name=f'channel_range{i}',
                               parameter_class=TraceParameter,
                               label=f'Range channel {i}',
                               unit='V',
                               initial_value=0.4,
                               val_mapping={0.4: 7})
            self.add_parameter(name=f'impedance{i}',
                               parameter_class=TraceParameter,
                               label=f'Impedance channel {i}',
                               unit='Ohm',
                               initial_value=50,
                               val_mapping={50: 2})

            self.add_parameter(name=f'bwlimit{i}',
                               parameter_class=TraceParameter,
                               label=f'Bandwidth limit channel {i}',
                               unit=None,
                               initial_value='DISABLED',
                               val_mapping={'DISABLED': 0,
                                            'ENABLED': 1})

        self.add_parameter(name='trigger_operation',
                           parameter_class=TraceParameter,
                           label='Trigger Operation',
                           unit=None,
                           initial_value='TRIG_ENGINE_OP_J',
                           val_mapping={'TRIG_ENGINE_OP_J': 0,
                                        'TRIG_ENGINE_OP_K': 1,
                                        'TRIG_ENGINE_OP_J_OR_K': 2,
                                        'TRIG_ENGINE_OP_J_AND_K': 3,
                                        'TRIG_ENGINE_OP_J_XOR_K': 4,
                                        'TRIG_ENGINE_OP_J_AND_NOT_K': 5,
                                        'TRIG_ENGINE_OP_NOT_J_AND_K': 6})
        n_trigger_engines = 2

        for i in range(1, n_trigger_engines+1):
            self.add_parameter(name=f'trigger_engine{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Engine {i}',
                               unit=None,
                               initial_value='TRIG_ENGINE_' + ('J' if i == 1 else 'K'),
                               val_mapping={'TRIG_ENGINE_J': 0,
                                            'TRIG_ENGINE_K': 1})
            self.add_parameter(name=f'trigger_source{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Source {i}',
                               unit=None,
                               initial_value='EXTERNAL',
                               val_mapping={'CHANNEL_A': 0,
                                            'CHANNEL_B': 1,
                                            'EXTERNAL': 2,
                                            'DISABLE': 3})
            self.add_parameter(name=f'trigger_slope{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Slope {i}',
                               unit=None,
                               initial_value='TRIG_SLOPE_POSITIVE',
                               val_mapping={'TRIG_SLOPE_POSITIVE': 1,
                                            'TRIG_SLOPE_NEGATIVE': 2})
            self.add_parameter(name=f'trigger_level{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Level {i}',
                               unit=None,
                               initial_value=140,
                               vals=validators.Ints(0, 255))

        self.add_parameter(name='external_trigger_coupling',
                           parameter_class=TraceParameter,
                           label='External Trigger Coupling',
                           unit=None,
                           initial_value='DC',
                           val_mapping={'AC': 1, 'DC': 2})
        self.add_parameter(name='external_trigger_range',
                           parameter_class=TraceParameter,
                           label='External Trigger Range',
                           unit=None,
                           initial_value='ETR_2V5',
                           val_mapping={'ETR_TTL': 2, 'ETR_2V5': 3})
        self.add_parameter(name='trigger_delay',
                           parameter_class=TraceParameter,
                           label='Trigger Delay',
                           unit='Sample clock cycles',
                           initial_value=0,
                           vals=validators.Multiples(divisor=8, min_value=0))
        # See Table 3 - Trigger Delay Alignment
        # TODO: this is either 8 or 16 dependent on the  number of channels in use

        # NOTE: The board will wait for a for this amount of time for a
        # trigger event.  If a trigger event does not arrive, then the
        # board will automatically trigger. Set the trigger timeout value
        # to 0 to force the board to wait forever for a trigger event.
        #
        # IMPORTANT: The trigger timeout value should be set to zero after
        # appropriate trigger parameters have been determined, otherwise
        # the board may trigger if the timeout interval expires before a
        # hardware trigger event arrives.
        self.add_parameter(name='timeout_ticks',
                           parameter_class=TraceParameter,
                           label='Timeout Ticks',
                           unit='10 us',
                           initial_value=0,
                           vals=validators.Ints(min_value=0))

        self.add_parameter(name='aux_io_mode',
                           parameter_class=TraceParameter,
                           label='AUX I/O Mode',
                           unit=None,
                           initial_value='AUX_IN_AUXILIARY',
                           val_mapping={'AUX_OUT_TRIGGER': 0,
                                        'AUX_IN_TRIGGER_ENABLE': 1,
                                        'AUX_IN_AUXILIARY': 13})
        self.add_parameter(name='aux_io_param',
                           parameter_class=TraceParameter,
                           label='AUX I/O Param',
                           unit=None,
                           initial_value='NONE',
                           val_mapping={'NONE': 0,
                                        'TRIG_SLOPE_POSITIVE': 1,
                                        'TRIG_SLOPE_NEGATIVE': 2})

        # ----- Parameters for the acquire function -----
        self.add_parameter(name='mode',
                           label='Acquisition mode',
                           unit=None,
                           initial_value='NPT',
                           set_cmd=None,
                           val_mapping={'NPT': 0x200, 'TS': 0x400})
        self.add_parameter(name='samples_per_record',
                           label='Samples per Record',
                           unit=None,
                           initial_value=1024,
                           set_cmd=None,
                           vals=validators.Multiples(
                                divisor=self.samples_divisor, min_value=256))
        self.add_parameter(name='records_per_buffer',
                           label='Records per Buffer',
                           unit=None,
                           initial_value=10,
                           set_cmd=None,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='buffers_per_acquisition',
                           label='Buffers per Acquisition',
                           unit=None,
                           set_cmd=None,
                           initial_value=10,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='channel_selection',
                           label='Channel Selection',
                           unit=None,
                           set_cmd=None,
                           initial_value='AB',
                           val_mapping={'A': 1, 'B': 2, 'AB': 3})
        self.add_parameter(name='transfer_offset',
                           label='Transfer Offset',
                           unit='Samples',
                           set_cmd=None,
                           initial_value=0,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='external_startcapture',
                           label='External Startcapture',
                           unit=None,
                           set_cmd=None,
                           initial_value='ENABLED',
                           val_mapping={'DISABLED': 0X0,
                                        'ENABLED': 0x1})
        self.add_parameter(name='enable_record_headers',
                           label='Enable Record Headers',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x8})
        self.add_parameter(name='alloc_buffers',
                           label='Alloc Buffers',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x20})
        self.add_parameter(name='fifo_only_streaming',
                           label='Fifo Only Streaming',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x800})
        self.add_parameter(name='interleave_samples',
                           label='Interleave Samples',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x1000})
        self.add_parameter(name='get_processed_data',
                           label='Get Processed Data',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x2000})
        self.add_parameter(name='allocated_buffers',
                           label='Allocated Buffers',
                           unit=None,
                           set_cmd=None,
                           initial_value=4,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='buffer_timeout',
                           label='Buffer Timeout',
                           unit='ms',
                           set_cmd=None,
                           initial_value=1000,
                           vals=validators.Ints(min_value=0))

        self.add_parameter(name='trigger_holdoff',
                           label='Trigger Holdoff',
                           docstring=f'If enabled Alazar will '
                                     f'ignore any additional triggers '
                                     f'while capturing a record. If disabled '
                                     f'this will result in corrupt data. '
                                     f'Support for this requires at least '
                                     f'firmware version '
                                     f'{self._trigger_holdoff_min_fw_version}',
                           vals=validators.Bool(),
                           get_cmd=self._get_trigger_holdoff,
                           set_cmd=self._set_trigger_holdoff)

        model = self.get_idn()['model']
        if model != 'ATS9373':
            raise Exception(f"The Alazar board kind is not 'ATS9373',"
                            f" found '{str(model)}' instead.")

    def _get_trigger_holdoff(self) -> bool:
        fwversion = self.get_idn()['firmware']
        if not isinstance(fwversion, str) or LooseVersion(fwversion) < \
                LooseVersion(self._trigger_holdoff_min_fw_version):
            return False

        # we want to check if the 26h bit (zero indexed) is high or not
        output = np.uint32(self._read_register(58))
        # the two first two chars in the bit string is the sign and a 'b'
        # remove those to only get the bit pattern
        bitmask = bin(output)[2:]
        # all prefixed zeros are ignored in the bit conversion so the
        # bit mask may be shorter than what we expect. in that case
        # the bit we care about is zero so we return False
        if len(bitmask) < 27:
            return False

        return bool(bin(output)[-27])

    def _set_trigger_holdoff(self, value: bool) -> None:
        fwversion = self.get_idn()['firmware']
        if not isinstance(fwversion, str) or LooseVersion(fwversion) < \
                LooseVersion(self._trigger_holdoff_min_fw_version):
            raise RuntimeError(f"Alazar 9373 requires at least firmware "
                               f"version {self._trigger_holdoff_min_fw_version}"
                               f" for trigger holdoff support. "
                               f"You have version {fwversion}")
        current_value = self._read_register(58)

        if value is True:
            # to enable trigger hold off we want to flip the
            # 26th bit to 1. We do that by making a bitwise or
            # with a number that has a 1 on the 26th place and zero
            # otherwise. We use numpy.unit32 instead of python numbers
            # to have unsigned ints of the right size
            enable_mask = np.uint32(1 << 26)
            new_value = current_value | enable_mask
        else:
            # to disable trigger hold off we want to flip the
            # 26th bit to 0. We do that by making a bitwise and
            # with a number that has a 0 on the 26th place and 1
            # otherwise
            disable_mask = ~np.uint32(1 << 26)
            new_value = current_value & disable_mask
        self._write_register(58, int(new_value))



from typing import Any

from qcodes import VisaInstrument
from qcodes.utils.helpers import create_on_off_val_mapping
from qcodes.utils.validators import Enum, Strings


class Keithley_2400(VisaInstrument):
    """
    QCoDeS driver for the Keithley 2400 voltage source.
    """
    def __init__(
            self,
            name: str,
            address: str,
            **kwargs: Any):
        super().__init__(name, address, terminator='\n', **kwargs)

        self.add_parameter('rangev',
                           get_cmd='SENS:VOLT:RANG?',
                           get_parser=float,
                           set_cmd='SOUR:VOLT:RANG {:f}',
                           label='Voltage range')

        self.add_parameter('rangei',
                           get_cmd='SENS:CURR:RANG?',
                           get_parser=float,
                           set_cmd='SOUR:CURR:RANG {:f}',
                           label='Current range')

        self.add_parameter('compliancev',
                           get_cmd='SENS:VOLT:PROT?',
                           get_parser=float,
                           set_cmd='SENS:VOLT:PROT {:f}',
                           label='Voltage Compliance')

        self.add_parameter('compliancei',
                           get_cmd='SENS:CURR:PROT?',
                           get_parser=float,
                           set_cmd='SENS:CURR:PROT {:f}',
                           label='Current Compliance')

        self.add_parameter('volt',
                           get_cmd=self._get_read_output_protected,
                           get_parser=self._volt_parser,
                           set_cmd=':SOUR:VOLT:LEV {:.8f}',
                           label='Voltage',
                           unit='V',
                           docstring="Sets voltage in 'VOLT' mode. "
                                     "Get returns measured voltage if "
                                     "sensing 'VOLT' otherwise it returns "
                                     "setpoint value. "
                                     "Note that it is an error to read voltage with "
                                     "output off")

        self.add_parameter('curr',
                           get_cmd=self._get_read_output_protected,
                           get_parser=self._curr_parser,
                           set_cmd=':SOUR:CURR:LEV {:.8f}',
                           label='Current',
                           unit='A',
                           docstring="Sets current in 'CURR' mode. "
                                     "Get returns measured current if "
                                     "sensing 'CURR' otherwise it returns "
                                     "setpoint value. "
                                     "Note that it is an error to read current with "
                                     "output off")

        self.add_parameter('mode',
                           vals=Enum('VOLT', 'CURR'),
                           get_cmd=':SOUR:FUNC?',
                           set_cmd=self._set_mode_and_sense,
                           label='Mode')

        self.add_parameter('sense',
                           vals=Strings(),
                           get_cmd=':SENS:FUNC?',
                           set_cmd=':SENS:FUNC "{:s}"',
                           label='Sense mode')

        self.add_parameter(
            'output',
            set_cmd=':OUTP:STAT {}',
            get_cmd=':OUTP:STAT?',
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0")
        )

        self.add_parameter('nplcv',
                           get_cmd='SENS:VOLT:NPLC?',
                           get_parser=float,
                           set_cmd='SENS:VOLT:NPLC {:f}',
                           label='Voltage integration time')

        self.add_parameter('nplci',
                           get_cmd='SENS:CURR:NPLC?',
                           get_parser=float,
                           set_cmd='SENS:CURR:NPLC {:f}',
                           label='Current integration time')

        self.add_parameter('resistance',
                           get_cmd=self._get_read_output_protected,
                           get_parser=self._resistance_parser,
                           label='Resistance',
                           unit='Ohm',
                           docstring="Measure resistance from current and voltage "
                                     "Note that it is an error to read current "
                                     "and voltage with output off")

        self.write(':TRIG:COUN 1;:FORM:ELEM VOLT,CURR')
        # This line sends 2 commands to the instrument:
        # ":TRIG:COUN 1" sets the trigger count to 1 so that each READ? returns
        # only 1 measurement result.
        # ":FORM:ELEM VOLT,CURR" sets the output string formatting of the the
        # Keithley 2400 to return "{voltage}, {current}".
        # Default value on instrument reset is "VOLT, CURR, RES, TIME, STATUS";
        # however, resistance, status, and time are unused in this driver and
        # so are omitted.
        # These commands do not reset the instrument but do the minimal amount
        # to ensure that voltage and current parameters can be read from the
        # instrument, in the event that output formatting of the instrument was
        # previously changed to some other unknown state.
        self.connect_message()

    def _get_read_output_protected(self) -> str:
        """
        This wrapper function around ":READ?" exists because calling
        ":READ?" on an instrument with output disabled is an error.
        So first we check that output is on and if not we return
        nan for volt, curr etc.
        """
        output = self.output.get_latest()
        if output is None:
            # if get_latest returns None we have
            # to ask the instrument for the status of output
            output = self.output.get()

        if output == 1:
            msg = self.ask(':READ?')
        else:
            raise RuntimeError("Cannot perform read with output off")
        return msg

    def _set_mode_and_sense(self, msg: str) -> None:
        # This helps set the correct read out curr/volt
        if msg == 'VOLT':
            self.sense('CURR')
        elif msg == 'CURR':
            self.sense('VOLT')
        else:
            raise AttributeError('Mode does not exist')
        self.write(f':SOUR:FUNC {msg:s}')

    def reset(self) -> None:
        """
        Reset the instrument. When the instrument is reset, it performs the
        following actions.

            Returns the SourceMeter to the GPIB default conditions.

            Cancels all pending commands.

            Cancels all previously send `*OPC` and `*OPC?`
        """
        self.write(':*RST')

    def _volt_parser(self, msg: str) -> float:
        fields = [float(x) for x in msg.split(',')]
        return fields[0]

    def _curr_parser(self, msg: str) -> float:
        fields = [float(x) for x in msg.split(',')]
        return fields[1]

    def _resistance_parser(self, msg: str) -> float:
        fields = [float(x) for x in msg.split(',')]
        res = fields[0] / fields[1]
        return res
